/*
	action_queue: workqueue for some worker thread

	part of DerMixD
	(c)2007/8 Thomas Orgis, licensed under GPLv2
*/
#include "common.hxx"
#include "action_queue.hxx"

action_queue::action_queue()
{
	pthread_mutex_init(&lock, NULL);
	sem_init(&sem, 0, 0);
	fill = 0;
}

action_queue::~action_queue()
{
	/* If there are still actions left, delete them. */
	if(fill > 0)
	{
		action* act;
		while( (act=pull()) != NULL ) delete act;
	}
	sem_destroy(&sem);
	pthread_mutex_destroy(&lock);
}

void action_queue::push(action *act)
{
	pthread_mutex_lock(&lock);
	q.push(act);
	++fill;
	sem_post(&sem);
	pthread_mutex_unlock(&lock);
}

action* action_queue::pull()
{
	action* ret;
	if(safe_sem_wait(&sem) != 0)
	{
		serror("Ugh! Unable to wait for a semaphore!");
		return NULL;
	}
	// Now, after waiting, get the first item.
	pthread_mutex_lock(&lock);
	if(fill < 1) ret = NULL;
	else
	{
		xdebug("poppin' work from list with size %zu", q.size());
		ret = q.front();
		q.pop();
		--fill;
		xdebug("poppin' work done from list with size %zu", q.size());
	}
	pthread_mutex_unlock(&lock);
	return ret;
}
