#ifndef DMD_H_AUDIO_FUNCTIONS
#define DMD_H_AUDIO_FUNCTIONS

/*
	audio_functions: audio data manipulation
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/


//clipper method out of mixplayer.c of mixplayd

#define MAX_SAMPLE  32766
#define CLIP        30000
#define CLIP_A      ((MAX_SAMPLE-CLIP)*(MAX_SAMPLE-CLIP))
#define CLIP_B      (MAX_SAMPLE-2*CLIP)

//return soft-clipped short-sized value for integer input value
inline audio_type clipper(int in)
{
	if((in < CLIP )&&( in >-CLIP)) return (audio_type) in;
	else if (in>0) return (audio_type) (MAX_SAMPLE-(CLIP_A/(CLIP_B+in)));
	else return (audio_type) (-(MAX_SAMPLE-(CLIP_A/(CLIP_B-in))));
}

// different resamplers (+amplifiers)
// multiple channels are most problematic here (cpu intensive)
void fullsample(audio_type* A, unsigned int a,  int* B, unsigned int b, double vol);
void fullsample_hq(audio_type* A, unsigned int a,  int* B, unsigned int b, double vol, int* startval);
void fullsample_hqbuff(audio_buffer& A, int* B, unsigned int b, double vol, int* startval);
void fullsample_hqbuff2(audio_buffer& A, mixer_buffer &B, unsigned int b, double vol, int* startval);
void fullchansample(audio_buffer& A, mixer_buffer& B, unsigned int b, double vol, int* startval);


/* calculate seconds from samples */
inline double seconds(volatile audio_format* format, unsigned long pos){ return (double)1.0*pos/format->rate; }
/* calculate samples from seconds */
inline unsigned long samples(volatile audio_format* format, double sec){ return (unsigned long) (sec * format->rate); }
/* calculate bytes from samples */
inline unsigned long bytes(volatile audio_format* format, unsigned long samples){ return samples*format->channels*sizeof(audio_type); }

#endif
