#ifndef DMD_H_COMMON
#define DMD_H_COMMON

/*
	common.hxx: common system header includes and some definitions
	
	part of DerMixD
	(c)2004-8 Thomas Orgis, licensed under GPLv2
*/

#include <pthread.h>
#include <cstdlib>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sched.h>
#include <semaphore.h>
#include <string>
#include <vector>
#include <queue>
#include <limits.h>
#include <errno.h>
// For hacky thread scheduling on Linux.
#ifdef LINUX
#include <sys/time.h>
#include <sys/resource.h>
#include <linux/unistd.h>
extern int   base_priority;
extern pid_t base_pid;
#endif
// The debug header _after_ the std header because of conflicts, esp. with the "construct" macro.
#include "debug.hxx"

using namespace std;

#ifdef WANT_PORT_T
typedef in_port_t porttype;
#else
typedef u_int16_t porttype;
#endif

/*
	audio format used for mixing...
	I have to take care if I want to use any other format than 44100Hz/16bit/stereo.
	Infrastructure for different formats is _partially_ there.
	Some code still assumes stereo; also audio_type is fixed as short (16bit at least on my x86... and alpha)
	
	Problem is the drop in performance for audio manipulations (additional loop level for channels)
	Possible option: have mono,stereo and multichannel versions of the important functions (definitely: fullsample)
	and the a switch(channels){ 1: mono() : 2: stereo() default: multi() }; with a reasonable chunk size the impact of this one should be minimal
	
	but changing the 16bit (short)... well...
*/

//values are set in main
extern unsigned long int audio_rate;
extern unsigned int audio_channels;
extern unsigned int buffer_samples;
//as I consider this standard for the environment
#include "param.hxx"
extern param_space param;

//many do need this...
#include "comm_data.hxx"

// hacks for that strange situation of a thread hanging in sem_wait() while it should be cancelled
#define sem_wait_cancel(sem) sem_wait(sem), pthread_testcancel()
int sem_cancel(pthread_t thread, sem_t *sem); // implementation in mixer_actions.cxx
// a really notorious thing... but the checking for EINTR is really necessary since signals occur (and be it ctrl+z!)
// ...I could place the testcancel in the loop code block to have the sem_wait_cancel behaviour... but it should'n be needed in any case
// wait until eith sem_wait is successful or a different error than EINTR occured
inline int safe_sem_wait(sem_t *sem){ int r = 0; do { r=sem_wait(sem); } while(r != 0 && errno == EINTR); pthread_testcancel(); return r; }
// do trywait until it's not successful anymore and the error is not EINTR
inline void sem_zero(sem_t *sem){ do { pthread_testcancel(); } while(sem_trywait(sem) == 0 || errno == EINTR); pthread_testcancel(); }

#endif
