#ifndef DMD_H_COMMUNICATOR
#define DMD_H_COMMUNICATOR

/*
	communicator: all the server stuff (port watcher and socket communicator)
	
	part of DerMixD
	(c)2004-8 Thomas Orgis, licensed under GPLv2
	
	depends on netcomm to provide actual socket reading/writing
	
	providing the function and the data structure to start the port watcher
	socketeers are fully internal...
	NOTE: There is only _one_ port_watcher() with its associated data, no object to create multiple instances of!
	This may stay that way until I somehow need to listen to more than one port or simply are bored enough to do it right just because it should be that way...
*/

#include "common.hxx"
#include "action.hxx"
#include "communicator_actions.hxx"
#include "thread_container.hxx"
#include "socket_writer_data.hxx"
#include "tstring.hxx"

//initialization data for port watcher

class watchers_data;

class watcher_data: public thread_container
{
public:
	thread_data td;
	pthread_t thread;
	int retval;
	socket_listener* sl;
	int sock;
	watchers_data *mum;

	watcher_data(watchers_data* mummy){ construct("watcher_data"); mum = mummy; };
	~watcher_data(){ destruct("watcher_data"); };
};

// Holds socketeer reaper thread and list of socket threads on watch.
// init() and exit() are _not_ thread-safe!
// Note that for some strange reason I opted to use explicit init() and exit() here instead of constructor/destructor.
class watchers_data: public thread_container
{
private:
	bool add_watcher(watcher_data *wd);
	bool good;
public:
	// Needed pointers for socketeer creation as well as thread reporting.
	// Will replace the copying through init() with something better in future?
	vector<action*> * actions;
	pthread_mutex_t * actions_lock;
	sem_t * semac;
	sem_t * reapsem; // Clear up that hack with reapsem_p!!

	// One reaper.
	thread_data td_reaper;
	pthread_t reapthread;
	// Multiple watchers (TCP, UNIX domain).
	vector<watcher_data*> watchers;
	// Copies pointers and creates reaper thread.
	bool init(vector<action*> *act, pthread_mutex_t *act_lock, sem_t *act_sem, sem_t *reap_sem);
	// Kills all watchers and the reaper.
	void exit(void);
	size_t thread_info(vector<tstring*> &tl);
	bool add_tcp(const porttype port, const int local);
	bool add_unix(const string name);
};

extern watchers_data wdat;

// For socketeer data... avoiding dynamic data in threads.
#define SCRATCH_CHARS 29

//data for a socketeer...
class socketeer_data: public thread_container
{
public:

	thread_data td_read, td_write;
	int sock;
	int retval;
	char *cwd;
	volatile int die_insane_bitch;
	string ms;
	string ans;
	action act; //this one for parsing
	action post_act; //this one for postaction
	action notify_act; //this one for notification (act still needs to be present)
	vector<action*> * actions;
	pthread_mutex_t * actions_lock;
	socket_buffer * sb;
	comm_data * cd;
	sem_t * semac;
	pthread_t thread;
	int erg;
	bool do_notify, do_post;
	const action_def* postaction; //does the current action imply another one (and which one) after success?
	char scratch[SCRATCH_CHARS+1]; //for ugly sprintf numbers

	socket_writer_data wrdat; //data for associated writer
	socketeer_data(int s, vector<action*> * ac, pthread_mutex_t * aclock, sem_t * sac);
	~socketeer_data();
	void prepare();
};


extern sem_t *reapsem_p; // HACK!
void reap_socketeer(pthread_t *deadite); // interface to reaper thread
void reap_all_socketeers(); // to be called before exit, without reaper thread

#endif
