#part of ThOrMA (but not necessarily depending upon)
#DerMixD::Control: interface to dermixd with some added logic

#Copyright (c) 2005/6 Thomas Orgis (thomas@orgis.org)
#This is Open Source, Distribution and Modification under the terms of the Artistic License
#see the file LICENSE in the ThOrMA package root or at a place some steps upwards directory-wise if you got this file in a separate package


package DerMixD::Control;

use strict;
use DerMixD::Base;
use List::Util qw(reduce);

our @ISA = ('DerMixD::Base');
my $me = 'DerMixD::Control';


sub Control
{
	#example: $pal->Control(seek,20)
	my $self = shift;
	my $coms = shift;
	my $ret = '';
	#the intelligent part of control is the insertion of active channel when needed...
	#needed when: single start/pause/stop
	#or seek/load/play/bass/mid/treble/speed <arg>
	#or eq <arg1> <arg2> <arg3>
	my $ac;
	my $channelquest;
	my $numberquest;
	if($#{$coms} == 0) #really single command or just channel missing
	{
		if(grep($_ eq $coms->[0], ('pause','stop','start')))
		{
			if(defined ($ac = $self->ActiveChannel()))
			{
				#more convenient, but not illogical
				$coms->[0] = 'start' if($coms->[0] eq 'pause' and $ac->{status} =~ 'pause');
				#just append channel number
				$coms->[1] = $ac->{num};
			}
			else{ $channelquest = 1; }
		}
	}
	elsif($#{$coms} == 1 or ($coms->[0] eq 'eq' and $#{$coms} == 3)) #insert channel
	{
		if(grep($_ eq $coms->[0], ('seek','rseek','eq','bass','mid','treble','speed','pitch','load','play','vol','volume')))
		{
			if(defined ($ac = $self->ActiveChannel()))
			{
				splice(@{$coms}, 1, 0, $ac->{num}); #put channel in place, interpret numbers later
			}
			else{ $channelquest = 1; }
		}
	}
	#mangle numbers
	if($#{$coms} == 2 and grep($_ eq $coms->[0],('seek','bass','mid','treble','speed','volume','vol')))
	{
		#interpret numbers
		if($coms->[0] eq 'seek')
		{
			$numberquest = 1 unless defined ($coms->[2] = $self->AbsoluteNumber($coms->[2],$ac->{position},$ac->{length}));
		}
		else
		{
			if($coms->[0] eq 'vol'){ $coms->[2] /= 100 unless $coms->[2] =~ '%'; $coms->[0] = 'volume'; }
			$numberquest = 1 unless defined ($coms->[2] = $self->AbsoluteNumber($coms->[2],$ac->{$coms->[0]},$ac->{$coms->[0]}));
		}
	}
	
	return $channelquest ? ['error: What is the active channel?'] : ($numberquest ? ['error: Unable to interpret number!'] : $self->Command($coms));
}

#guess the channel that is the current active one
#again with some respect for apparent fading...
#still not smooth...
sub ActiveChannel
{
	my $self = shift;
	$self->FullStatus();
	if(@{$self->{in}})
	{
		my $actor = $self->{in}->[0]; #default is first channel
		#find all channels that have a file loaded
		if(my @candidates = grep( ($_->{file} ne '' and $_->{length} > 0) , @{$self->{in}}))
		{
			#now find players or if pausing one if there are none...
			my @list = grep( $_->{status} =~ 'play', @candidates); #playing and seek-play
			my $fading = $#list > 0 ? 1 : 0;
			@list = grep( $_->{status} =~ 'pause', @candidates) unless @list; #paused and seek-pause
			@list = grep( $_->{status} eq 'stopped', @candidates) unless @list; #...
			$actor = reduce { $b->{position}/$b->{length} < $a->{position}/$a->{length} ? $b : $a } @list ;
			$actor->{fading} = $fading;
		}
		return $actor;
	}
	return undef;
}

sub AbsoluteNumber
{
	my ($self, $rel, $now, $ref) = @_;
	my $num = undef;
	if($rel =~ /^([+-]{0,1})(\d*\.\d+|\d+)(%{0,1})$/)
	{
		if($3){ $num = $2/100*$ref; } #percents
		else{ $num = $2; }  #normal units
		if($1) #offset, relative
		{
			my $factor = ($1 eq '+') ? 1 : -1;
			$num = $now + $factor*$num;
		}
	}
	elsif($rel =~ /^([+-]{0,1})(\d+):((\d*\.\d+|\d+)|(\d+):(\d*\.\d+|\d+))$/)
	{
		if($6 ne ''){ $num = $2*3600+$5*60+$6; }
		else{ $num = $2*60+$4; }
		if($1) #offset, relative
		{
			my $factor = ($1 eq '+') ? 1 : -1;
			$num = $now + $factor*$num;
		}
	}
	return $num;
}

sub Wait
{
	my $self = shift;
	my $arg = shift;
	return 0 unless($#{$arg} >= 1);
	$arg->[2] = 'wakeup' unless defined $arg->[2];
	my $res = $self->Command(['script',$arg->[0],$arg->[1],'say',$arg->[2]]);
	return $res->[0] =~ /^\s*error\:/ ? 0 : $self->WaitForSay($arg->[2]);
}

return 1;
