/*
	in_dummy: dummy input device
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "audio_functions.hxx"
#include "input.hxx"
#include "in_dummy.hxx"

#define ME "dummy_input"
const string dummy_input::t = "dummy";

dummy_input::dummy_input(input_data* ind)
{
	construct(ME);
	data = ind;
	type = &t;
	data->format.rate = 44100;
	data->format.channels = 2;
}

dummy_input::~dummy_input()
{
	destruct(ME);
}

int dummy_input::load(comm_data* cd, const string& filename) 
{
	debug("loading file %s", filename.c_str());
	data->position = 0;
	data->at_end = false;
	data->length = 4410000; // 100s
	data->filename = filename;
	sem_post(cd->waitress);
	return 0;
}

//really dummy
//on any real read think of always getting the numbers for all channels in a block!!!
int dummy_input::play(unsigned int n_samples)
{
	unsigned int to_fill = n_samples;

	data->position += to_fill;

	if(data->position > data->length)
	{
		sdebug("the end");
		to_fill = n_samples-(data->position - data->length);
		data->position = data->length;
		data->at_end = true;
	}

	unsigned int i;

	debug("filling buffer with %u samples", to_fill);
	if( to_fill > buffer.size)
	{
		if( buffer.reallocate(n_samples) == 0 ){ serror("error on reallocating memory"); return -1; }
	}

	//read

	buffer.fill = to_fill;

	xdebug("filling buffer at %p with %u samples with channelcount of %u", (void*) buffer.ptr, buffer.fill, buffer.channels);
	for(i = 0; i < buffer.fill; ++i){ buffer.ptr[2*i] = buffer.ptr[2*i+1] = 0; }

	sxdebug("filled");

	sem_post(data->insem);
	sxdebug("posted");

	return 0;
}



int dummy_input::sampleseek(comm_data* cd, unsigned long position)
{
	debug("seeking to %lu", position);

	if(position < data->length) { data->position = position; data->at_end = false; }
	else{ data->position = data->length; data->at_end = true; }

	if(cd != NULL)
	{
		cd->errcode = 0;
		cd->retval = seconds(&(data->format), position);
		sem_post(cd->waitress);
	}
	return 0;
}

int dummy_input::seek(comm_data* cd, double position) //seek to sample
{
	unsigned long samplepos = samples(&(data->format), position);
	return sampleseek(cd, samplepos);
}

//need real eq...
int dummy_input::set_eq(comm_data* cd)
{
	if(cd != NULL) sem_post(cd->waitress);
	return 0;
}
