/*
	main: the outer shell
	
	part of DerMixD
	(c)2004-2006 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "version.hxx"
#include "mixer.hxx"
#include "param_init.hxx"

#include "action.hxx"
#include "communicator_actions.hxx"

param_space param;

//these are _not_ volatile since only the main thread will use them... it will see the changes made by itself, won't it? 
unsigned long int audio_rate = 0; //set via param
unsigned int audio_channels = 2;
unsigned int buffer_samples = 0; //set via param
#ifdef LINUX
int   base_priority = 0;
pid_t base_pid = 0;
#endif

bool globalflag_oldstyle = false;

int main(int argc, char** argv)
{
	param_init();

	bool daemon = true;
	int only_local = 1;
	int staninit = 1;
	bool goodcommand = true;
	bool give_help = false;
	bool give_api = false;

#ifdef LINUX
	base_pid      = getpid();
	base_priority = getpriority(PRIO_PROCESS, 0); // Could do error check via errno...
#endif
	if(argc >= 2)
	{
		string arg;
		for(int i = 1; i < argc; ++i)
		{
			arg = argv[i];
			if( (arg == "-c") || (arg == "--console") ) daemon = false;
			else if( (arg == "-r") || (arg == "--remote") ) only_local = 0;
			else if( (arg == "-n") || (arg == "--nothing") ) staninit = 0;
			else if( (arg == "-O") || (arg == "--oldstyle") ) globalflag_oldstyle = true;
			else if( (arg == "-h") || (arg == "--help") ) give_help = 1;
			else if( arg == "--show-api") give_api = 1;
			else if( arg == "--version" ){ cout << "DerMixD v" << version << "." << subversion << endl; exit(0); }
			else if( (arg == "-b") || (arg == "--buffer") )
			{
				if(++i >= argc){ cerr << "Missing argument to option b/buffer (buffer size)!" << endl; goodcommand = false; }
				else param.set("main","buffer",argv[i]);
			}
			else if( (arg == "-a") || (arg == "--samplerate") )
			{
				if(++i >= argc){ cerr << "Missing argument to option a/samplerate (the rate...)!" << endl; goodcommand = false; }
				param.set("main","audio_rate",argv[i]);
			}
			else if( (arg == "-p") || (arg == "--port") )
			{
				if(++i < argc) param.set("main","port",argv[i]);
				else
				{
					cerr << "--port needs the port number as argument!" << endl;
					goodcommand = false; 
				}
			}
			else if( (arg == "-o") || (arg == "--output") )
			{
				if(++i == argc){ cerr << "--output needs a filename as argument!" << endl; goodcommand = false; }
				else param.set("main","output",argv[i]);
				//daemon_output = argv[i];
			}
			else
			{
				if(!param.parse(arg))
				{
					cerr << "Unknown switch / parameter setting " << arg << " given!" << endl;
					goodcommand = false;
				}
			}
		}
	}
	
	if(give_help)
	{
		cout << "DerMixD" << endl << " - a tcp-controlled audio file mixing daemon in the tradition of mixplayd (well, on the outside) - " << endl;
		cout << "Version " << version << "." << subversion << ",  Written and thus copyrighted in " << years << " by Thomas Orgis. Licensed under the GPL 2. No warranty of any kind. Have fun;-)" << endl;
		cout << endl;
		cout << "Usage: dermixd [-c|--console] [-r|--remote] [-h|--help] [other parameters]" << endl;
		cout << "" << endl << "Some configuration parameters can be given as \"-o\" for simple on/off options and \"-p <value>\" or \"--param <value>\" for parameters with values:" << endl;
		cout << "-c, --console\tThis disables daemon mode (stay in foreground, then)." << endl;
		cout << "-r, --remote\tAllow connections from the outside. Default is to only accept connections from localhost. Consider that there is no authorization checking for clients, be it local or remote (so, at least  every user on your machine can talk to me)!" << endl;
		cout << "-p, --port\tspecify the TCP port to use" << endl;
		cout << "-n, --nothing\tStart with no initial setup (otherwise: 2 inputs -> 1 default output)" << endl;
		cout << "-b, --buffer <positive_number>\tspecify mixer buffer size in samples" << endl;
		cout << "-a, --samplerate <positive_number>\tspecify mixer and output sAmple rAte" << endl;
		cout << "-o, --output <filename>\tin daemon mode stdout/err get redirected to that file (debugging)" << endl;
		cout << "-O, --oldstyle\tuse the old - and deprecated! - practice of [say] and [watch] strings without command prefix" << endl;
		cout << "--version\tjust print name and version" << endl;
		cout << "--show-api\tprint out all control commands for use by clients (the API)" << endl;
		cout << "-h, --help\tgive this help info" << endl;
		cout << endl << "And there are the more complete configuration parameters organized in groups you can specify as \"group.name=<value>\" on the command name or in the future via configuration file:" << endl;
		param.help(stdout);
		exit(0);
	}
	else if(give_api)
	{
		cout << "Full API listing:" << endl << endl;
		vector<string*> ah;
		action_allhelp(external_api, ah);
		
		for(vector<string*>::iterator i = ah.begin(); i != ah.end(); ++i)
		cout << **i << endl;
		
		exit(0);
	}
	
	if(!goodcommand){ cerr << "Please read the help (switch -h)" << endl; exit(1); }
	
	buffer_samples = param.as_uint("main","buffer");
	if(buffer_samples < 2){ cerr << "minimum buffer size is 2 (and you don't expect pitch shifting to work in that area, do you?)!" << endl; exit(2); }
	
	audio_rate = param.as_uint("main","audio_rate");
	if(audio_rate == 0){ cerr << "no... please give an audio rate above zero!" << endl; exit(3); }
	
	audio_channels = param.as_uint("main","channels");
	if(!((audio_channels == 1) || (audio_channels == 2))){ cerr << "no... only mono or stereo mode supported!" << endl; exit(4); }

	if(!param.as_bool("tcp") && !param.as_bool("unix"))
	{
		serror("Please be reasonable and let me communicate via TCP or UNIX domain socket!");
		return 100;
	}

	return mixer(daemon, only_local, staninit);
}
