/*
	mixer: the mixer, the master, the gate, the keeper
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
//#include <algorithm>

#include <fcntl.h>
#include <signal.h>

#include "audio_buffer.hxx"
#include "audio_functions.hxx"

#include "output.hxx"
#include "input.hxx"

#include "netcomm.hxx"
#include "socket_writer_data.hxx"
#include "action.hxx"
#include "communicator.hxx"
#include "mixer_data.hxx"

#include "mixer_tools.hxx"
#include "mixer_actions.hxx"

#include "secsleep.hxx"

//ok... some global variables... yes, I'll do it "properly" some time...
#include "mixer_stuff.hxx"
unsigned char globalstate = GLOBAL_RUNNING;
vector<inchannel*> inners;
vector<outchannel*> outers;
sem_t outsem, insem, semac, reapsem;
vector<action*> actions;
vector<channel*> scripters;
pthread_mutex_t playlock = PTHREAD_MUTEX_INITIALIZER;

int mixer(bool become_daemon, int only_local, int staninit)
{
	debug("[mixer] startup with pid %i", (int)getpid());
	porttype port = (porttype) param.as_uint("main","port");

	if(sem_init(&outsem, 0 , 0) != 0) 
	{
		serror("[mixer] Cannot initialize semaphore!");
		exit(1);
	}

	if(sem_init(&insem, 0 , 0) != 0) 
	{
		serror("[mixer] Cannot initialize semaphore!");
		exit(1);
	}

	if(sem_init(&semac, 0 , 0) != 0) 
	{
		serror("[mixer] Cannot initialize semaphore!");
		exit(1);
	}
	if(sem_init(&reapsem, 0 , 0) != 0)
	{
		serror("[mixer] Cannot initialize semaphore!");
		exit(1);
	}
	reapsem_p = &reapsem;

	audio_format format;
	pthread_mutex_t actions_lock = PTHREAD_MUTEX_INITIALIZER;
	unsigned int i;

	//pthread_mutex_t inlock = PTHREAD_MUTEX_INITIALIZER, outlock = PTHREAD_MUTEX_INITIALIZER;
	unsigned int active_inputs = 0;
	unsigned int active_outputs = 0;

	format.rate = audio_rate;
	format.channels = audio_channels;

	vector<inchannel*>::iterator ini;
	vector<outchannel*>::iterator outi;
	vector<action*> script;

	string namestring;
	int ec;

	sdebug("[mixer] before daemon decision");
	if(become_daemon)
	{
		string daemon_output;
		param.get("main","output",daemon_output);
		
		cerr << "Becoming a daemon and starting the communication server on port " << port << " then...\n";

		int pid = fork();

		if (pid < 0) exit(1); /* error encountered, no child has been created! */
		if (pid != 0) exit(0); /* this is the parent, and hence should be terminated */

		setsid(); /* make the process a group leader, session leader, and lose control tty */

		/* close STDOUT, STDIN, STDERR */

  	close(STDIN_FILENO);
  	close(STDOUT_FILENO);
  	close(STDERR_FILENO);

  	/* reopen STDOUT, STDIN, STDERR */
    if (open("/dev/null", O_RDONLY) == -1) exit(2);
    if (open(daemon_output.c_str(), O_WRONLY|O_CREAT|O_TRUNC, 0666) == -1) exit(3);
    if (open(daemon_output.c_str(), O_WRONLY|O_CREAT|O_APPEND, 0666) == -1) exit(4);

  	signal(SIGHUP, SIG_IGN); /* ignore SIGHUP that will be sent to a child of the process */

  	chdir("/"); /* change to working dir */

  	pid = fork();

		if (pid < 0) exit(1); /* fork() failed, no child process was created! */
		if (pid != 0) exit(0); /* this is the parent, hence should exit */

  	/* this is the child process of the child process of the actual calling process */
  	/* and can safely be called a grandchild of the original process */ 
	}

	/* ignore SIGPIPE, for reading, writing to non-opened pipes */
	/* every program using pipes should ignore this signal for being on the safe side */
	signal(SIGPIPE, SIG_IGN);

	//standard setup _after_ being daemon beccause otherwise threads of output/inputs get lost
	if(staninit)
	{
		string default_output;
		string default_outfile;
		param.get("main","default_output",default_output);
		param.get("main","default_outfile",default_outfile);

		cerr << "Creating initial setup: 2 input channels both serving 1 " << default_output << " output.\n";

		namestring = "Helena";
		cout << default_output << " output (" << namestring << "): ";
		
		if((ec=add_output(outers, namestring)) == 0) cout << "OK\n";
		else 
		{
			error("[mixer] Error code %i! Aborting...", ec);
			exit(1);
		}

		namestring = default_outfile;
		comm_data* tempcd = new comm_data(&outsem); 
		if(tempcd == NULL){ serror("Oooohh..."); exit(5); }
		outers.back()->load(default_output,namestring,tempcd);
		safe_sem_wait(&outsem);
		if(tempcd->errcode){ error("Error %i on opening %s!",tempcd->errcode, namestring.c_str()); exit(3); }
		delete tempcd;

		if(outers.back()->device->activate())	outers.back()->status = PLAYING;
		else{ serror("Cannot activate playback!"); exit(3); }

		namestring = "Lena";
		cout << "Input 0 (" << namestring << "): ";
		if( (ec=add_input(inners, namestring, insem)) == 0) cout << "OK\n";
		else 
		{
			cout << "Error code " << ec << "!\nAborting...\n";
			exit(1);
		}

		namestring = "Elena";
		cout << "Input 1 (" << namestring << "): ";
		if( (ec=add_input(inners, namestring, insem)) == 0) cout << "OK\n";
		else 
		{
			error("[mixer] Error code %i! Aborting...", ec);
			exit(1);
		}

		cout << "Wow... now let's bind the input channels to the output.\n";

		cout << "Bind Lena: ";
		if( (ec=bind_input(inners[0], outers[0])) == 0) cout << "OK\n";
		else 
		{
			error("[mixer] Error code %i! Aborting...", ec);
			exit(1);
		}

		cout << "Bind Elena: ";
		if( (ec=bind_input(inners[1], outers[0])) == 0) cout << "OK\n";
		else 
		{
			error("[mixer] Error code %i! Aborting...", ec);
			exit(1);
		}

		cerr << "That went fine... I'll conclude the initial setup with a full listing:\n";
		vector<string*> cst;
		set_stats(inners,outers,&cst);
		for(vector<string*>::iterator i = cst.begin(); i != cst.end(); ++i)
		{
			printf("%s\n", (*i)->c_str());
			delete(*i);
		}
	}

	//start com server with it's reaper
	signal(SIGPIPE, SIG_IGN);
	wdat.init(&actions, &actions_lock, &semac, &reapsem);
	if(param.as_bool("tcp"))
	if(!wdat.add_tcp(port, only_local))
	{
		error("Unable to listen on TCP socket %i!", (int)port);
		exit(5);
	}

	if(param.as_bool("unix"))
	if(!wdat.add_unix(param("socket")))
	{
		error("Unable to listen on UNIX domain socket %s!", param("socket").c_str());
		exit(5);
	}

	//operation
	#ifdef DEBUG_LEAK2
	globalstate = GLOBAL_DEAD;
	#endif
	while(!globalstate) //while  == GLOBAL_RUNNING
	{

		//stage1: admin, timed scripting now there!

		//only if actions are there
		// If this call is interrupted by a signal (thus not sucessful), the actions just will be a bit delayed.
		if(sem_trywait(&semac) == 0)
		{
			pthread_mutex_lock(&actions_lock);
			sdebug("[mixer] executing actions");
			process_actions(actions);
			sdebug("[mixer] actions done");
			sem_zero(&semac);
			pthread_mutex_unlock(&actions_lock);
			sdebug("[mixer] lock released");
		}

		//now execute script actions... if there
		//note that one can also have just added script actions that get executed now
		if(!scripters.empty())
		{
			for(vector<channel*>::iterator c = scripters.begin(); c != scripters.end(); ++c)
			{
				#ifdef XDEBUG_SCRIPT
				double n,t;
				(*c)->now_and_then(buffer_samples, n, t);
				cerr << "searching script action (now=" << n << " then=" << t << ")" << endl;
				#endif
				if((*c)->status == PLAYING)
				{
					//getactions fills any current actions into script and returns true if there are still actions left
					if(! (*c)->getactions(buffer_samples,script))
					{
						scripters.erase(c);
						--c;
					}
				}
			}
			if(!script.empty()) //found an action
			{
				#ifdef DEBUG_SCRIPT
				cerr << "[main] found script actions\n";
				#endif
				//process and finally delete the actions
				process_actions(script);
				#ifdef DEBUG_SCRIPT
				cerr << "[main] done with script actions\n";
				#endif
			}
		}
		
		if(globalstate == GLOBAL_PAUSED)
		{
			//take a rest and wait for the doorbell
			sdebug("[mixer] sleeping and waiting for some action (devices still open!)...");
			safe_sem_wait(&semac);
			sem_post(&semac); //put the value back for really getting the actions
			sdebug("[mixer] good morning - got some action...");
			globalstate = GLOBAL_RUNNING;
		}
		else if(globalstate == GLOBAL_DEAD) break; //could delete all inputs and outputs in actions instead...
		
		//stage2a: beginning to fetch input data for upcoming chunk
		active_inputs = 0;
		pthread_mutex_lock(&playlock); // Make sure scanners don't interfere.
		for(ini = inners.begin(); ini != inners.end(); ++ini)
		{
			//play(number) means start reading of <number> samples into device buffer, this is sizeof(audio_type)*in.format.channels bytes!!!
			if( !(*ini)->working && (((*ini)->status == PLAYING) || (*ini)->buffering) )
			{	
			 	unsigned int need_samples = (unsigned int) ((*ini)->speed*(buffer_samples - ((*ini)->buffer.fill)) * ((*ini)->data.format.rate) / format.rate);
				if(need_samples)
				{
					xdebug("[mixer] ordering %u samples (speed %g) on ch %i %s", need_samples, (*ini)->speed, (*ini)->data.id, (( (*ini)->buffering ) ? " for buffering" : " for playback"));
					(*ini)->device->play(need_samples);
					(*ini)->ordered_samples = need_samples;
					++active_inputs;
				}
				else
				{
					if(((*ini)->status == PLAYING) && (*ini)->ordered_samples)
					{
						sdebug("[mixer] prebuffered input now active");
						++active_inputs;
						// trick it a bit...
						sem_post(&insem);
					}
					else
					{
						error("[mixer] Eek! channel %i wants to play/buffer nuthin! stopping it.", (*ini)->data.id);
						(*ini)->buffering = false;
						(*ini)->status = STOPPED;
					}
				}
			}
		}

		if(active_inputs)
		{
			//yet another idea: give a big buffer to the output that gets normal memcpy / addition from the inbuffers and only one clipping stage (which may perform better with big mixes) ... in the normal single channel playback this wouldn't do anything good, though. One could optimize the process for many cases (not just) like crazy. Be careful before penetrating the flexible method that is here now and does the job as well as good old mixplayd!

			//stage2b: wait for input data of upcoming chunk
			for(i = 0; i < active_inputs; ++i){ safe_sem_wait(&insem); } 		
			
			//stage3: mixing
			for(ini = inners.begin(); ini != inners.end(); ++ini)
			{
				if(!(*ini)->working && ((*ini)->status == PLAYING || (*ini)->buffering) )
				{
					xdebug("[mixer] ic %i buffering/playing at position %lu", (*ini)->data.id, (*ini)->data.position);
					//we want this number of input samples
					unsigned int news = (unsigned int) (1.0 * ((*ini)->device->buffer.fill) / ((*ini)->ordered_samples) * (buffer_samples - (*ini)->buffer.fill));

					//resample + amplify from device buffer to inchannel buffer
					//simple
					//fullsample((audio_type*) (*ini)->device->buffer.ptr, (*ini)->device->buffer.fill, (*ini)->buffer, news, (*ini)->volume);
					//more quality?
					//fullsample_hq((audio_type*) (*ini)->device->buffer.ptr, (*ini)->device->buffer.fill, (*ini)->buffer.ptr+2*(*ini)->buffer.fill, news, (*ini)->volume, (*ini)->startval);
					//(*ini)->buffer.fill += news;
					//channel conversion inclusive
					fullchansample((*ini)->device->buffer, (*ini)->buffer, news, (*ini)->volume, (*ini)->startval);
					xdebug("[mixer] have: %u", (*ini)->buffer.fill);

					//basically mixing...
					if( (*ini)->status == PLAYING )
					{
						//two steps: 
						//	- check if track at end and if another channel should follow seamlessly
						//	- mixing
						//end?
						unsigned int a;
						if( ((*ini)->speed != 0) && (((*ini)->buffer.fill < buffer_samples) || ((*ini)->data.at_end)) ) //both indicate track end, but only when the title has some playing speed
						{
							debug("[mixer] track on ch %i ended (position: %lu / %lu; fill: %u / %u; at_end: %i)", (*ini)->data.id, (*ini)->data.position, (*ini)->data.length, (*ini)->buffer.fill, buffer_samples, (int)(*ini)->data.at_end);
							//stop...
							(*ini)->status = STOPPED;
							(*ini)->device->sampleseek(NULL,0); //note: the actual seek will take place only on subsequent playback
							if(!(*ini)->postscript.empty())
							{
								sdebug("[mixer] adding postactions to worklist");
								pthread_mutex_lock(&actions_lock);
								(*ini)->getpostactions(actions);
								sem_post(&semac);
								pthread_mutex_unlock(&actions_lock);
								sdebug("[mixer] lock released");
							}
							//check device sanity and delete if necessary
							//the end could have occured just because of some decoder trouble
							(*ini)->device_sane(); 

							if( (*ini)->follower != NULL )
							{
								sdebug("[mixer] there is a follower");
								//try to get data from follower's buffer to fill up my buffer
								news = (buffer_samples - (*ini)->buffer.fill);
								if( (*ini)->follower->buffer.fill >=  news)
								{
									/*
										avoid hard clicks; p.ex. because volumes don't match
										even when the tracks fit seamlessly in principle,
										there	can be compression artefacts and a different volume
										
										a bit of linear smoothing should do it
										
										be careful with overflowing/clipping!
									*/
									if(audio_channels == 2)
									{
										audio_type gap[2];
										int range = (news > 50) ? 50 : news;
										int rapos = range+1;
										gap[0] = (*ini)->startval[0] - (*ini)->follower->buffer.ptr[0];
										gap[0] += (gap[0] < 0) ? 1 : ((gap[0] == 0) ? 0 : -1);
										gap[1] = (*ini)->startval[1] - (*ini)->follower->buffer.ptr[1];
										gap[1] += (gap[1] < 0) ? 1 : ((gap[1] == 0) ? 0 : -1);
										#ifdef DEBUG_SMOOTHBUFFER
										cerr << "gaps: " << gap[0] << " and " << gap[1] << "; first title having " << (*ini)->buffer.fill << " samples; startvals" << (*ini)->startval[0] << " and " << (*ini)->startval[1] << endl;
										#endif
										for(a = 0; a < news*2; a = a+2)
										{
											(*ini)->buffer.ptr[a + (*ini)->buffer.fill*2] = (*ini)->follower->buffer.ptr[a];
											(*ini)->buffer.ptr[a + 1 + (*ini)->buffer.fill*2] = (*ini)->follower->buffer.ptr[a+1];
											if(--rapos > 0)
											{
												#ifdef DEBUG_SMOOTHBUFFER
												cerr << "smoothing " << rapos << " at sample " << a/2+(*ini)->buffer.fill << " from (" << (*ini)->buffer.ptr[a + (*ini)->buffer.fill*2] << "|" << (*ini)->buffer.ptr[a + 1 + (*ini)->buffer.fill*2] << ")";
												#endif
												(*ini)->buffer.ptr[a + (*ini)->buffer.fill*2] += (audio_type)((float) rapos/range*gap[0]);
												(*ini)->buffer.ptr[a + 1 + (*ini)->buffer.fill*2] += (audio_type)((float) rapos/range*gap[1]);
												#ifdef DEBUG_SMOOTHBUFFER
												 cerr << " to (" << (*ini)->buffer.ptr[a + (*ini)->buffer.fill*2] << "|" << (*ini)->buffer.ptr[a + 1 + (*ini)->buffer.fill*2] << ")" << endl;
												#endif
											}
										}

										(*ini)->buffer.fill += news;
										#ifdef DEBUG_SMOOTHBUFFER
										(*ini)->buffer.printout();
										#endif
									}
									else
									{
										audio_type gap;
										int range = (news > 50) ? 50 : news;
										int rapos = range+1;
										gap = (*ini)->startval[0] - (*ini)->follower->buffer.ptr[0];
										gap += (gap < 0) ? 1 : ((gap == 0) ? 0 : -1);
										#ifdef DEBUG_SMOOTHBUFFER
										cerr << "gap: " << gap << "; first title having " << (*ini)->buffer.fill << " samples; startval" << (*ini)->startval[0] << endl;
										#endif
										for(a = 0; a < news; ++a)
										{
											(*ini)->buffer.ptr[a + (*ini)->buffer.fill] = (*ini)->follower->buffer.ptr[a];
											if(--rapos > 0)
											{
												#ifdef DEBUG_SMOOTHBUFFER
												cerr << "smoothing " << rapos << " at sample " << a+(*ini)->buffer.fill << " from (" << (*ini)->buffer.ptr[a + (*ini)->buffer.fill] << ")";
												#endif
												(*ini)->buffer.ptr[a + (*ini)->buffer.fill] += (audio_type)((float) rapos/range*gap);
												#ifdef DEBUG_SMOOTHBUFFER
												 cerr << " to (" << (*ini)->buffer.ptr[a + (*ini)->buffer.fill] << ")" << endl;
												#endif
											}
										}

										(*ini)->buffer.fill += news;
										#ifdef DEBUG_SMOOTHBUFFER
										(*ini)->buffer.printout();
										#endif
									}
									
									//now shift follower's buffer
									memmove((*ini)->follower->buffer.ptr,
									        (*ini)->follower->buffer.ptr+news,
									        sizeof(audio_type)
									        * (*ini)->follower->buffer.channels
									        * ((*ini)->follower->buffer.fill - news));
									(*ini)->follower->buffer.fill -= news;
									debug("[mixer] added follower at %u", (*ini)->buffer.fill-news);
									#ifdef DEBUG_FOLLOW
									//(*ini)->buffer.printout();
									//(*ini)->follower->buffer.printout();
									#endif
								}
								//the follower is following... LATER!
								action* mixact = new action; // deleted by process_actions
								mixact->def = &in_start;
								mixact->ints.clear();
								mixact->ints.push_back((*ini)->follower->data.id);
								pthread_mutex_lock(&actions_lock);
								actions.push_back(mixact);
								pthread_mutex_unlock(&actions_lock);
								sem_post(&semac);
								debug("deleting leader %i from %i as well as follower %i from %i", (*ini)->follower->leader->data.id, (*ini)->follower->data.id, (*ini)->follower->data.id, (*ini)->data.id );
								(*ini)->follower->leader = NULL;
								(*ini)->follower = NULL; //and this entry has fulfilled its quest
							}
						}

						xdebug("[mixer] going to mix for ic %i", (*ini)->data.id);
						//the actual mixing stage
						for(outi = (*ini)->targets.begin(); outi != (*ini)->targets.end(); ++outi)
						{
							if( (*outi)->status == PLAYING )
							{
								//now take ini's buffer and add it to outi's buffer...
								//with a bit of clipping prevention
								if((*outi)->device->buffer.virgin)
								{
									//first values, simply copy (should save time)
									for(a = 0; a < (*ini)->buffer.fill*audio_channels; ++a)
									{
										(*outi)->device->buffer.ptr[a] = clipper( (*ini)->buffer.ptr[a] );
									}
									(*outi)->device->buffer.virgin = false;
								}
								else
								{
									//mixing: now add and declip again
									for(a = 0; a < (*ini)->buffer.fill*audio_channels; ++a)
									{
										(*outi)->device->buffer.ptr[a] = clipper( (*ini)->buffer.ptr[a] + (*outi)->device->buffer.ptr[a] );
									}
								} 
								//the output buffer is possibly not fully filled
								if((*outi)->device->buffer.fill < (*ini)->buffer.fill) (*outi)->device->buffer.fill = (*ini)->buffer.fill;
							}
						}

						sxdebug("[mixer] mixing done");
						(*ini)->buffer.fill = 0;
						if( !(*ini)->watchers.empty() ) telltime(*ini);	
					}
					else { (*ini)->buffering = false; } 
				}
			}		
		}
		pthread_mutex_unlock(&playlock); // Now the input scanners can do I/O again...

		//stage4: activate output; double buffering being inherent now with threads and waiting for last chunk inside the device object

		/*
			Babble about open issues:

			All PLAYING outputs are played, even when there is no actual data there.
			The resulting silent output (with no music) keeps the device
			silently busy and prevents any on/off sounds that are not desired or 
			some nasty buffer from keeping a chunk of audio until playback starts again.
			Heard that. Not nice.

			When you want a device to really not be busy then set the channel to PAUSING
			(device still open but no output) or SUSPEND (keep channel around but close
			audio device).
			
			With the new output_device-level multithreading it should be possible for the one output needing silence to push pending audio really out before pauses - OSS - to issue the writing of some silent buffers (post the semaphore multiple times; but be careful with the play() side of waiting)
		*/

		active_outputs = 0;
		for(outi = outers.begin(); outi != outers.end(); ++outi)
		{
			if((*outi)->status == PLAYING)
			//if((*outi)->device->online())
			{
				++active_outputs;
				//xdebug("[mixer] playing device %i, called %s", (*outi)->data.id, (*outi)->name.c_str());
				//play silence if nothing else - ugly workaround! to be handled by output_basic / oss_output itself in future!
				if((*outi)->device->buffer.fill == 0) (*outi)->device->buffer.fill = (*outi)->device->buffer.size;
				(*outi)->device->play();
			}
		}

		//some precautions for unemployment
		if(active_outputs == 0)
		{
			if(active_inputs == 0)
			{
				//really nothing to do: no input, no output
				//wait for some action
				//socketeers trigger semac on every mixer command
				sdebug("[mixer] waiting for some action...");
				safe_sem_wait(&semac);
				sem_post(&semac); //put the value back for really getting the actions
				sdebug("[mixer] got some action...");
			}
			else
			{
				//no output, but inputs running... that means decode/read stuff as fast as possible into the void or wait a bit here to allow user to react to this quite senseless state without struggling with a very busy cpu...
				//the delay time 
				sxdebug("[mixer] just sleeping a bit...");
				//usleep((useconds_t) ((float) buffer_samples / audio_rate * 900000));
				secsleep((double) buffer_samples / audio_rate * 0.9);
			}
		}
		else while(--active_outputs){}
	} //le big end

	//destroy it all

	sdebug("[mixer] going berserk");

	// Reap socket watchers and socketeer reaper.
	wdat.exit();
	// reap the remaining socketeers
	reap_all_socketeers();

	sdebug("[mixer] cancelled & joined the port / socket stuff...");

	//delete input devices
	for(ini = inners.begin(); ini != inners.end(); ++ini)
	{
		debug("[mixer] would like to delete %p", *ini);
		delete (*ini);
		debug("[mixer] indeed deleted %p", *ini);
	}
	inners.clear();

	sdebug("[mixer] cleared inners");	

	//delete output devices
	for(outi = outers.begin(); outi != outers.end(); ++outi)
	{
		debug("[mixer] would like to delete %p", *outi);
		delete (*outi);
		debug("[mixer] indeed deleted %p", *outi);
	}

	sdebug("[mixer] clearing outers");
	outers.clear();
	sdebug("[mixer] cleared outers");	

	sem_destroy(&outsem); 
	sem_destroy(&insem); 
	sem_destroy(&semac); 
	sem_destroy(&reapsem);

	if(!become_daemon) cerr << "finished normally" << endl;

	return 0;
}
