/*
	mixer_actions: outsourced code for mixer's daily work
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"

#include "audio_buffer.hxx"
#include "audio_fifo.hxx"
#include "audio_functions.hxx"

#include "input.hxx"
#include "output.hxx"

#include "netcomm.hxx"
#include "socket_writer_data.hxx"
#include "action.hxx"
#include "communicator.hxx"
#include "mixer_data.hxx"
#include "mixer_tools.hxx"
#include "mixer_actions.hxx"
#include "mixer_stuff.hxx"
#include "tstring.hxx"

extern bool globalflag_oldstyle;

//some of the actions of the mixer, mainly these needed outside the main loop

//emulate getstat from mixplayd
//not sure how long this code will stay
void set_compat_stats(	vector<inchannel*> &inners, vector<outchannel*> &outers, vector<string*> *compat_stats)
{
	tstring ts;
	vector<inchannel*>::iterator ini;
	bool o = true;
	int n = -1;
	for(ini = inners.begin(); ini != inners.end(); ++ini)
	{
		int minsec[2], sec[2];
		minsec[0] = (int) seconds(&(*ini)->data.format, (*ini)->data.position);
		sec[0]    = minsec[0] - (minsec[0] / 60) * 60;
		minsec[1] = (int) seconds(&(*ini)->data.format, (*ini)->data.length);
		sec[1]    = minsec[1] - (minsec[1] / 60) * 60;
		if(o) ts="00:00:00";
		ts.sprintf(" ch %i %s %02d:%02d..%02d:%02d (%i) [%0.2f]",
		            ++n, (compatstatus((*ini)->status)).c_str(),
		            minsec[0]/60, sec[0], minsec[1]/60, sec[1],
		            (int) ((*ini)->volume*100), (*ini)->speed     );
		if(!o)
		{
			compat_stats->push_back(new string(ts));
			ts = "";
			o = true;
		}
		else o = false;
	}
	if(!o)
	{
		compat_stats->push_back(new string(ts));
	}
}

//my way of status, there should be another function for just checking the playback status of one channel (mixplayer.pl always calls getstat for this task -> we have push info with watch now
void set_stats(	vector<inchannel*> &inners, vector<outchannel*> &outers, vector<string*> *stats)
{
	vector<inchannel*>::iterator ini;
	vector<outchannel*>::iterator outi;
	tstring ts;

	stats->push_back(new string("[fullstat] +begin"));
	for(ini = inners.begin(); ini != inners.end(); ++ini)
	{
		ts.sprintf("in %i:", (*ini)->data.id);
		ts += (*ini)->name;
		ts += " -> (";
		for(outi = (*ini)->targets.begin(); outi != (*ini)->targets.end(); ++outi)
		{
			if(outi > (*ini)->targets.begin()) ts.sprintf(",%i", (*outi)->data.id);
			else ts.sprintf("%i", (*outi)->data.id);
		}
		ts += ") ";
		ts += stringstatus((*ini)->status);
		ts.sprintf(" at %gs/%gs (buf: %lu) vol: %g speed: %g eq: %g|%g|%g",
		            seconds(&((*ini)->data.format), (*ini)->data.position),
		            seconds(&((*ini)->data.format), (*ini)->data.length),
		            (unsigned long)(*ini)->buffer.fill, (*ini)->volume, (*ini)->speed,
		            (*ini)->data.bass, (*ini)->data.mid, (*ini)->data.treble );
		//leader, script and device/file
		if((*ini)->leader != NULL) ts.sprintf(", following %i:%s", (*ini)->leader->data.id, (*ini)->leader->name.c_str());
		if( !(*ini)->script.empty() ) ts.sprintf(", %lu actions", (unsigned long)(*ini)->script.size());
		if((*ini)->device != NULL)
		{
			ts += ", dev: ";
			ts += *(*ini)->device->type;
			ts += " file: ";
			ts += (*ini)->data.filename;
		}
		stats->push_back(new string(ts));
		ts = "";
	}
	for(outi = outers.begin(); outi != outers.end(); ++outi)
	{
		ts.sprintf("out %i:", (*outi)->data.id);
		ts += (*outi)->name + " <- (";
		for(ini = (*outi)->sources.begin(); ini != (*outi)->sources.end(); ++ini)
		{
			if(ini > (*outi)->sources.begin()) ts.sprintf(",%i", (*ini)->data.id);
			else ts.sprintf("%i", (*ini)->data.id);
		}
		ts += ") " + stringstatus((*outi)->status);
		if( !(*outi)->script.empty() ) ts.sprintf(", %lu actions", (unsigned long)(*outi)->script.size());
		if((*outi)->device != NULL)
		{
			ts += " dev: ";
			ts += (*outi)->device->type();
			ts += " file: ";
			ts += (*outi)->data.filename;
		}

		stats->push_back(new string(ts));
		ts = "";
	}
	stats->push_back(new string("[fullstat] -end"));
}

unsigned char add_output(vector<outchannel*> &outers, string &name)
{
	unsigned char errcode = 0;
	outers.push_back(new outchannel((int) outers.size(), name));
	return errcode;
}

//add a new input channel
//simple, eh?
unsigned char add_input(vector<inchannel*> &inners, string &name, sem_t &insem)
{
	unsigned char errcode = 0;
	inners.push_back(new inchannel((int) inners.size(), name, &insem));
	if(inners.back()->status == DEAD)
	{
		debug("[mixer] creation of inchannel %i failed", (int) inners.size()-1);
		inners.pop_back();
		errcode = 1;
	}
	return errcode;
}

void tellsomething(inchannel* inch, const string &msg)
{
	vector<socket_writer_data*>::iterator swi;
	tstring ts; // Better temporary copy explicitly here _once_.
	ts.sprintf("[ch%i] ", inch->data.id);
	ts += msg;
	for(swi = inch->watchers.begin(); swi != inch->watchers.end(); ++swi)
	(*swi)->write_copy(ts);
}

//0: ok
//1: is busy (loading, seeking), try later (to prevent other clients from hanging)!
//2: wrong channel number
unsigned char remove_input(vector<inchannel*> &inners, vector<outchannel*> &outers,long &num)
{
	vector<inchannel*>::iterator ini;
	vector<outchannel*>::iterator outi;
	unsigned char errcode = 0;
	if( (num >= 0) && (num < (int) inners.size()) )
	{ 
		if( (inners[num]->status == LOADING) || (inners[num]->status == SEEKING_PLAY) || (inners[num]->status == SEEKING_PAUSE))
		{
			errcode = 1;
		}
		else
		{
			//delete from all targets' sources
			for(outi = inners[num]->targets.begin(); outi != inners[num]->targets.end(); ++outi)
			{
				//now deleting from outi's sources list
				unbind_source(inners[num], *outi);
			}
			//set all following ids to match their future position
			int num2 = num;
			for(ini = inners.begin() + num + 1; ini != inners.end(); ++ini)
			{
				debug("[mixer] removing input %li ... inchannel %i is becoming %i", num, (*ini)->data.id, num2);
				(*ini)->data.id = num2++;
			}
			//finally, remove it
			delete inners[num];
			inners.erase(inners.begin() + num);
		}
	}
	else
	{
		//bad channel number
		errcode = 2;
	}
	return errcode;
}

//the same for output channels
unsigned char remove_output(vector<inchannel*> &inners, vector<outchannel*> &outers,long &num)
{
	vector<inchannel*>::iterator ini;
	vector<outchannel*>::iterator outi;
	unsigned char errcode = 0;
	if( (num >= 0) && (num < (int) outers.size()) )
	{
		if( outers[num]->status == LOADING ) errcode = 1;
		else
		{
			//delete from all sources' targets
			for(ini = outers[num]->sources.begin(); ini != outers[num]->sources.end(); ++ini)
			{
				//now deleting from ini's targets list
				unbind_target(*ini, outers[num]);
			}
			//set all following ids to match their future position
			int num2 = num;
			for(outi = outers.begin() + num + 1; outi != outers.end(); ++outi)
			{
				debug("[mixer] deleting output %li... outchannel %i is becoming %i", num, (*outi)->data.id, num2);
				(*outi)->data.id = num2++;
			}
			//finally, remove it
			delete outers[num];
			outers.erase(outers.begin() + num);
		}
	}
	else
	{
		//bad channel number
		errcode = 2;
	}
	return errcode;
}

//binding an input channel to an output channel
unsigned char bind_input(inchannel* in, outchannel* out)
{
	bool bound = false;
	/* later: check for binding multiple times... or may this be wanted? */
	for(vector<outchannel*>::iterator outi = in->targets.begin(); outi != in->targets.end(); ++outi)
	{
		if(*outi == out){ bound = true; break; }
	}
	if(!bound)
	{
		in->targets.push_back(out);
		out->sources.push_back(in);
		if(!in->watchers.empty()) tellsomething(in, tstring().sprintf("bind %i", out->data.id));
	}
	return 0;
}

//add swr to the list of sockets that get playback info
void in_watch(inchannel* inch, socket_writer_data* swr)
{
	vector<socket_writer_data*>::iterator swi;
	for(swi = inch->watchers.begin(); swi != inch->watchers.end(); ++swi)
	if(*swi == swr) break;
	// Now only push it if not found in the list.
	if(swi == inch->watchers.end()) inch->watchers.push_back(swr);
}

//remove swr from watchers, carefully deleting _all_ occurences
void in_unwatch(inchannel* inch, socket_writer_data* swr)
{
	vector<socket_writer_data*>::iterator swi;
	for(swi = inch->watchers.begin(); swi != inch->watchers.end(); ++swi)
	{
		//only one entry is legal...
		if(*swi == swr){ inch->watchers.erase(swi); break; }
	}
}

void unwatch_all(vector<inchannel*> &inners, socket_writer_data* swr)
{
	vector<inchannel*>::iterator ini;
	for(ini = inners.begin(); ini != inners.end(); ++ini) in_unwatch(*ini, swr);
}

//this really is a general channel method...
void telltime(inchannel* inch)
{
	vector<socket_writer_data*>::iterator swi;
	//we have the last told time in inch->hadseconds
	//get the current secs by inch->position_in_seconds()
	//if different, print and save
	unsigned int ntime = inch->position_in_seconds();
	if(ntime != inch->hadseconds || inch->status != inch->hadstatus)
	{
		tstring ts;
		inch->hadseconds = ntime;
		inch->hadstatus = inch->status;
		if(globalflag_oldstyle) ts.sprintf("ch%i: %us, ", inch->data.id, ntime);
		else ts.sprintf("[ch%i] %us, ", inch->data.id, ntime);

		ts += stringstatus(inch->status);
		for(swi = inch->watchers.begin(); swi != inch->watchers.end(); ++swi)
		(*swi)->write_copy(ts);
	}
}

//delete inch from outch's source list
void unbind_source(inchannel* inch, outchannel* outch)
{
	debug("[mixer] deleting inchannel %i from sources of outchannel %i", inch->data.id, outch->data.id);
	vector<inchannel*>::iterator ini;
	for(ini = outch->sources.begin(); ini != outch->sources.end(); ++ini)
	{
		//ini is iterator of inchannel*, *ini is inchannel*, when == the channel to be deleted, delete it
		//only one entry is legal...
		if(*ini == inch){ outch->sources.erase(ini); break; }
	}
}

//delete outch from inch's target list
void unbind_target(inchannel* inch, outchannel* outch)
{
	debug("[mixer] deleting outchannel %i from targets of inchannel %i", outch->data.id, inch->data.id);
	vector<outchannel*>::iterator outi;
	for(outi = inch->targets.begin(); outi != inch->targets.end(); ++outi)
	{
		//only one entry is legal...
		if(*outi == outch){ inch->targets.erase(outi); break; }
	}
	tellsomething(inch, tstring().sprintf("unbind %i %i", outch->data.id));
}


//let's play: master and servant
void process_actions(vector<action*> &acts)
{
	vector<action*>::iterator k;
	for(k = acts.begin(); k != acts.end(); ++k)
	{
		debug("[mixer] action loop, act=%p def=%p comdat=%p", (*k), (*k)->def, (*k)->comdat);
		unsigned char errcode = 0;
		double retval = 0;
		string errtext;

		// one of these two will likely vanish
		bool feedback = false;
		bool wake_socketeer = false;
		if((*k)->comdat != NULL)
		{
			wake_socketeer = feedback = true;
			//preserve existing values... not clean...
			retval = (*k)->comdat->retval;
			errtext = (*k)->comdat->errtext;
		}

		unsigned int aid = (*k)->def->id;
		int flags = (*k)->def->flags;
		debug("[mixer] executing action type %u", aid);

		long num  = -1;
		long num2 = -1;
		string ms;

		if(flags & SOCK_IN)
		{
			num = ((*k)->def->id != NOTIFY) ? (*k)->ints[0] : (*k)->subaction->ints[0];
			if( (num >= 0) && (num < (long) inners.size()) )
			{
				inchannel* in = inners[num];
				switch( aid )
				{
					case NOTIFY:
						switch((*k)->subaction->def->id)
						{
							case IN_LOAD:
							case IN_INLOAD:
								in->working = false;
								if(in->offline) debug("[mixer] inch %li is just offline", num);
								else if(in->device_sane())
								{
									in->status = (*k)->comdat->errcode ? IDLE : STOPPED;
									errcode = (*k)->comdat->errcode ? LOADERR : 0;
									debug("[mixer] load on channel %li finished (success or failure)", num);
								}
								else
								{
									error("[mixer] load on channel %li failed for some reason (no sane device)", num);
									in->nodev();
									errcode = DEVERR;
								}
								if(!in->watchers.empty()) telltime(in);
							break;
							case IN_SEEK:
							case IN_RSEEK:
							case IN_LENGTH:
							case IN_SCAN:
								//if the seek actually failed but the device is still sane, we hope that we can resume playback...
								in->working = false;
								if(in->device_sane())
								{
									errcode = (*k)->comdat->errcode; // just forward it!
									if(errcode && errcode != NOT_SUPPORTED)
									{
										serror("Some error in seeking? Bailing out (seeks should fail gracefully).");
										in->nodev();
									}
									if(in->status == SEEKING_PAUSE) in->status = PAUSED;
									else if(in->status == SEEKING_PLAY) in->status = PLAYING;
									else errcode = 4;
								}
								else { in->nodev(); errcode = DEVERR; }
								if(!in->watchers.empty()) telltime(in);
							break;
							default: errcode = UNKNOWN_COMMAND;
						}
					break; // notify

					case IN_VOLUME:
						debug("[mixer] setting volume on inchannel %li to %g", num, (*k)->floats[0]);
						retval = in->volume = (*k)->floats[0];
						tellsomething(in, tstring().sprintf("volume %g", in->volume));
						//one could simply recalculate buffer here
						if(in->buffer.fill != 0) in->reset_buffer();
						errcode = 0;
					break;
					case IN_SPEED:
					case IN_PITCH:
						if((in->speed = aid==IN_SPEED ? (*k)->floats[0] : in->speed+(*k)->floats[0]) < 0) in->speed = 0;
						retval = in->speed;
						tellsomething(in, tstring().sprintf("speed %g", in->speed));
						debug("[mixer] setting speed on inchannel %li to %g", num, (*k)->floats[0]);
						if(in->buffer.fill != 0) in->reset_buffer();
						errcode = 0;
					break;

					case IN_BASS:
						debug("[mixer] setting bass on inchannel %li to %g", num, (*k)->floats[0]);
						retval = in->data.bass = (*k)->floats[0] > 0 ? (*k)->floats[0] : 0;
						tellsomething(in, tstring().sprintf("eq %g %g %g", in->data.bass, in->data.mid, in->data.treble));
						if(in->device != NULL)
						{
							if(in->device_sane())
							{
								(*k)->comdat->retval = retval;
								in->device->set_eq((*k)->comdat); wake_socketeer = false;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						if(in->buffer.fill != 0) in->reset_buffer();
					break;
					case IN_MID:
						debug("[mixer] setting mid on inchannel %li to %g", num, (*k)->floats[0]);
						retval = in->data.mid = (*k)->floats[0] > 0 ? (*k)->floats[0] : 0;
						tellsomething(in, tstring().sprintf("eq %g %g %g", in->data.bass, in->data.mid, in->data.treble));
						if(in->device != NULL)
						{
							if(in->device_sane())
							{
								(*k)->comdat->retval = retval;
								in->device->set_eq((*k)->comdat); wake_socketeer = false;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						if(in->buffer.fill != 0) in->reset_buffer();
					break;
					case IN_TREBLE:
						debug("[mixer] setting treble on inchannel %li to %g", num, (*k)->floats[0]);
						retval = in->data.treble = (*k)->floats[0] > 0 ? (*k)->floats[0] : 0;
						tellsomething(in, tstring().sprintf("eq %g %g %g", in->data.bass, in->data.mid, in->data.treble));
						if(in->device != NULL)
						{
							if(in->device_sane())
							{
								(*k)->comdat->retval = retval;
								in->device->set_eq((*k)->comdat); wake_socketeer = false;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						if(in->buffer.fill != 0) in->reset_buffer();
					break;
					case IN_EQ:
						errcode = 0;
						debug("[mixer] setting bass on inchannel %li to %g", num, (*k)->floats[0]);
						in->data.bass = (*k)->floats[0] > 0 ? (*k)->floats[0] : 0;
						debug("[mixer] setting mid on inchannel %li to %g", num, (*k)->floats[1]);
						in->data.mid = (*k)->floats[1] > 0 ? (*k)->floats[1] : 0;
						debug("[mixer] setting treble on inchannel %li to %g", num, (*k)->floats[2]);
						in->data.treble = (*k)->floats[2] > 0 ? (*k)->floats[2] : 0;
						tellsomething(in, tstring().sprintf("eq %g %g %g", in->data.bass, in->data.mid, in->data.treble));
						if(in->device != NULL)
						{
							if(in->device_sane())
							{
								in->device->set_eq((*k)->comdat); wake_socketeer = false;
								debug("[mixer] issued set_eq, wake_socketeer: %i", (int)wake_socketeer);
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						if(in->buffer.fill != 0) in->reset_buffer();
					break;
					case IN_START:
						debug("[mixer] wanna start inchannel %li", num);
						if(!in->working)
						{
							if(in->device_sane())
							{
								//basically we can start if there is a file loaded... even if it is already playing, the start should succeed, I guess... because we want that it is playing afterwards and this is the case, then.
								if( (in->status == STOPPED) || (in->status == PAUSED) || (in->status == PLAYING) )
								{
									in->status = PLAYING;
									errcode = 0;
								}
								else errcode = NONSENSE;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY; 
					break;
					
					case IN_STOP: //one should rewind here... 
						debug("[mixer] wanna stop inchannel %li", num);
						if(!in->working)
						{
							if(in->device_sane())
							{
								if( (in->status == PLAYING) || (in->status == PAUSED))
								{
									//This has been evil... it should actually work now with the delayed seekplay
									in->device->seek(NULL,0);
									if(in->buffer.fill != 0) in->reset_buffer(); //think about this and rewind!
									in->status = STOPPED;
									if(in->follower != NULL)
									{
										debug("[mixer] follower on stopped channel %i", in->data.id);
										in->follower->leader = NULL;
										in->follower->status = PLAYING;
										in->follower = NULL;
										sdebug("[mixer] activated it");
									}
									if(!in->watchers.empty()) telltime(in);
								}
								errcode = 0;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY; 
					break;
					case IN_PAUSE:
						if(!in->working)
						{
							if((in->status == PLAYING) || (in->status == STOPPED) || (in->status == PAUSED))
							{
								in->status = PAUSED;
								errcode = 0;
								if(!in->watchers.empty()) telltime(in);
							}
							else errcode = NONSENSE;
						}
						else errcode = BUSY; 
					break;
					case IN_LOAD: //determine appropriate device, create new if necessary and load the file
						debug("[mixer] wanna load file %s on inchannel %li", (*k)->strings[0].c_str(), num);
						if(!in->working)
						{
							if( in->load((*k)->strings[0], (*k)->comdat) == 0 )
							{
								sdebug("[mixer] load really going on...");
								wake_socketeer = false;
								in->working = true;
								if(!in->watchers.empty()) tellsomething(in, "load " + (*k)->strings[0]);
							}
							else errcode = NO_DEVICE;
						}
						else errcode = BUSY; 
					break;
					case IN_INLOAD: // load with predefined device
						debug("[mixer] wanna load file %s on inchannel %li", (*k)->strings[0].c_str(), num);
						if(!in->working)
						{
							if( in->load((*k)->strings[1], (*k)->comdat, (*k)->strings[0]) == 0)
							{
								sdebug("[mixer] load really going on...");
								wake_socketeer = false;
								in->working = true;
								if(!in->watchers.empty()) tellsomething(in, "load " + (*k)->strings[0]);
							}
							else errcode = NO_DEVICE;
						}
						else errcode = BUSY; 
					break;
					case IN_SEEK:
					case IN_RSEEK:
					case IN_LENGTH: // this is a self-controlled seek, of sorts
					case IN_SCAN:   // This, too.
						sdebug("[mixer] wanna seek/scan");
						if(!in->working)
						{
							if(in->device_sane())
							{
								if( (in->data.length > 0) && ((in->status == PLAYING) || (in->status == STOPPED) || (in->status == PAUSED)))
								{
									if(aid == IN_LENGTH)
									{
										in->device->scan_length((*k)->comdat);
									}
									else if(aid == IN_SCAN)
									{
										in->device->scan((*k)->comdat, (*k)->strings[0]);
									}
									else
									{
										double aimed_seconds = aid==IN_SEEK ? (*k)->floats[0] : seconds(&(in->data.format), in->data.position) + (*k)->floats[0];
										if(aimed_seconds < 0) aimed_seconds = 0;
										in->device->seek((*k)->comdat, aimed_seconds);
									}
									in->buffer.fill = 0; //just discard it now... maybe retain in future for length scan?
									sdebug("[mixer] triggered seek");
									if(feedback)
									{
										wake_socketeer = false;
										if(in->status == PLAYING) in->status = SEEKING_PLAY;
										else in->status = SEEKING_PAUSE;
										in->working = true;
									}
									if(!in->watchers.empty()) telltime(in);
								}
								else errcode = NONSENSE;
							}
							else { in->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY; 
					break;
					case IN_BIND:
						debug("[mixer] wanna bind inchannel %li to outchannel %li", num, num2);
						num2 = (*k)->ints[1]; //the outchannel to bind to
						if( ( num2 >= 0) && (num2 < (int) outers.size()) )errcode = bind_input(in,outers[num2]);
						else errcode = 3; //bad outchannel number
					break;

					case IN_UNBIND:
						num2 = (*k)->ints[1]; //the outchannel to unbind
						debug("[mixer] wanna unbind inchannel %li to outchannel %li", num, num2);
						if( ( num2 >= 0) && (num2 < (int) outers.size()) )
						{
							//delete outchannel from targets
							unbind_target(in,outers[num2]);
							//delete inchannel from sources
							unbind_source(in,outers[num2]);
							errcode = 0;
						}
						else errcode = BAD_CHANNEL;
					break;

					case IN_FOLLOW:
						num2 = (*k)->ints[1];
						if( (num2 >= 0) && (num2 < (int) inners.size()) && (inners[0]) )
						{
							//remove existing follower
							if(in->follower != NULL)
							{
								in->follower->leader = NULL;
								if(in->follower->buffer.fill != 0) in->follower->reset_buffer();
								tellsomething(in, tstring().sprintf("nofollow %i",in->follower->data.id));
								in->follower = NULL;
							}
							
							//prevent multiple following and invalid follower
							if(inners[num2]->leader != NULL) errcode = IN_FOLLOW_ANOTHER;
							else if(inners[num2]->device_sane() && (inners[num2]->data.length > 0))
							{
								//should it also be updated on further changes on in?
								in->follower = inners[num2];
								in->follower->buffering = true;
								in->follower->leader = in;
								tellsomething(in, tstring().sprintf("follow %i",in->follower->data.id));
								errcode = 0;
							}
							else errcode = IN_FOLLOW_INVALID;
						}
						else errcode = BAD_CHANNEL;
					break;
					
					case IN_NOFOLLOW:
					{
						debug("[mixer] removing follower of ch %i", in->data.id);
						if(in->follower != NULL)
						{
							debug("[mixer] ...which is ch %i", in->follower->data.id);
							in->follower->leader = NULL;
							if(in->follower->buffer.fill != 0) in->follower->reset_buffer();

							tellsomething(in, tstring().sprintf("nofollow %i",in->follower->data.id));
							in->follower = NULL;
						}
						errcode = 0;
					}
					break;

					case IN_BUFFER:
						if(!in->working)
						{
							if(in->device_sane())
							{
								if(in->status == PAUSED || in->status == STOPPED)
								in->buffering = true;
							}
							else{ in->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY;
					break;

					case IN_WATCH:
						in_watch(in, (*k)->swr);
					break;
					case IN_UNWATCH:
						in_unwatch(in, (*k)->swr);
					break;

					case IN_SCRIPT:
						//add script action to inchannel, checking channel validity (there must be a track loaded)
						{
							action* act = new action();
							*act = *(*k)->subaction; //this for sure won't work with a further subaction!
							debug("[mixer] adding script action at %p; copied from %p", (void*) act, (void*) (*k)->subaction);
							act->comdat = NULL; // swr stays there; not the direct thread comunication
							if((*k)->floats[0] < 0) in->addpostaction((*k)->floats[0], (*k)->ints[1], act);
							else
							if( in->addaction((*k)->floats[0], (*k)->ints[1], act) == 1 ) scripters.push_back((channel*) in);
						}
					break;
					
					case IN_CLEARSCRIPT:
						in->delscript();
						for(vector<channel*>::iterator sci = scripters.begin(); sci != scripters.end(); ++sci)
						{
							if((*sci) == (channel*) in){ scripters.erase(sci); break; }
						}
					break;

					case IN_SHOWSCRIPT:
					{
						sdebug("[mixer] showscript!");
						vector<string*>* vc = new vector<string*>;
						vc->push_back(new string);
						(*k)->begin(*vc->back());
						in->get_script_commands(*vc);
						vc->push_back(new string);
						(*k)->end(*vc->back());
						(*k)->swr->writes(vc);
					}
					break;
					
					default:
						errcode = UNKNOWN_COMMAND;
				}
			}
			else
			{
				//bad channel number
				errcode = BAD_CHANNEL;
			}
		}
		else if(flags & SOCK_OUT)
		{
			num = ((*k)->def->id != NOTIFY) ? (*k)->ints[0] : (*k)->subaction->ints[0];
			if( (num >= 0) && ( num < (long) outers.size()) )
			{
				outchannel* out = outers[num];
				switch( aid )
				{
					case OUT_START:
						debug("[mixer] wanna start outchannel %li", num);
						if(!out->working)
						{
							if(out->device_sane())
							{
								//basically we can start if there is a file loaded... even if it is already playing, the start should succeed, I guess... because we want that it is playing afterwards and this is the case, then.
								if( (out->status == STOPPED) || (out->status == PAUSED) || (out->status == PLAYING) )
								{
									if(out->device->activate())	out->status = PLAYING;
									else // just complain, but no status change - we could succeed later
									{
										errcode = DEVERR;
										errtext = "device failed to activate" + (out->device->problem.empty() ? "" : " <- " + out->device->problem );
									}
								}
								else errcode = NONSENSE;
							}
							else { out->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY; 
					break;

					case OUT_STOP: //one should rewind here... 
						debug("[mixer] wanna stop outchannel %li", num);
						if(!out->working)
						{
							if(out->device_sane())
							{
								if( (out->status == PLAYING) || (out->status == PAUSED))
								{
									out->status = STOPPED;
									out->device->deactivate();
								}
								errcode = 0;
							}
							else { out->nodev(); errcode = DEVERR; }
						}
						else errcode = BUSY; 
					break;

					case OUT_PAUSE:
						if(!out->working)
						{
							if((out->status == PLAYING) || (out->status == STOPPED) || (out->status == PAUSED))
							{
								out->status = PAUSED;
								errcode = 0;
							}
							else errcode = NONSENSE;
						}
						else errcode = BUSY; 
					break;

					case OUT_LOAD: //determine appropriate device, create new if necessary and load the file
						debug("[mixer] wanna load file %s (type %s) on outchannel %li", (*k)->strings[1].c_str(), (*k)->strings[0].c_str(), num);
						if(!out->working)
						{
							if( out->load((*k)->strings[0], (*k)->strings[1], (*k)->comdat) == 0 )
							{
								wake_socketeer = false;
								out->working = true;
							}
							else errcode = NO_DEVICE;
						}
						else errcode = BUSY; 
					break;

					case NOTIFY:
						switch((*k)->subaction->def->id)
						{
							case OUT_LOAD:
								out->working = false;
								if(out->offline) debug("[mixer] ch %li is just offline", num);
								else if(out->device_sane())
								{
									if(out->device->online())
									{
										//the comdat is always the same... one for one socketeer
										out->status = (*k)->comdat->errcode ? IDLE : STOPPED;
										errcode = (*k)->comdat->errcode ? LOADERR : 0;
										debug("[mixer] load on outchannel %li succeded", num);
									}
									else
									{
										errcode = LOADERR;
										out->status = IDLE;
										errtext = "device failed come up" + (out->device->problem.empty() ? "" : " <- " + out->device->problem);
									}
								}
								else
								{
									error("[mixer] load on outchannel %li failed for some reason (no sane device)", num);
									out->nodev();
									errcode = DEVERR;
								}
							break;
							default: errcode = UNKNOWN_COMMAND;
						}
					break;
					default:
						errcode = UNKNOWN_COMMAND;
				}
			}
			else
			{
				//bad channel number
				errcode = BAD_CHANNEL;
			}
		}
		else //general action
		{
			//could've been quit
			if(aid == QUIT)
			{
				globalstate = GLOBAL_DEAD;
				break;
			}
			else switch ( aid )
			{
				case GETSTAT:
					{
						sxdebug("[mixer] giving compat stats");
						//I don't have any error expected here... 
						vector<string*> *cst = new vector<string*>;
						set_compat_stats(inners, outers, cst);
						(*k)->swr->writes(cst);
					}
				break;
				//think of a way to compute this in the socketeer... without races, of course...
				//here it is way more safe
				case FULLSTAT:
					{
						//I used to have a bool to remember if I already have prepared the current stat_string, but how often will unnecessary double creation happen to justify this additional bloat of creating and initializing one bool...
						sxdebug("[mixer] giving stats");
						vector<string*> *cst = new vector<string*>;
						set_stats(inners, outers, cst);
						(*k)->swr->writes(cst);
					}
				break;
				case THREADSTAT:
					{
						// collect thread info from all input and output channels
						// You have to add the socket threads and "[threadstat] -end" later - I don't want to do the socketeer locking here.
						unsigned long count = 1; // mixer is one thread.
						string retl;
						(*k)->begin(retl);
						(*k)->comdat->retlines.push_back(new tstring(retl));
						(*k)->comdat->retlines.push_back(new tstring());
#ifdef LINUX
						(*k)->comdat->retlines.back()->sprintf("%i(%i @ %i): miXer", 0, (int)base_pid, base_priority);
#else
						(*k)->comdat->retlines.back()->sprintf("%i: miXer", 0);
#endif
						for(vector<inchannel*>::iterator  i = inners.begin(); i != inners.end(); ++i)
						count += (*i)->thread_info((*k)->comdat->retlines);

						for(vector<outchannel*>::iterator i = outers.begin(); i != outers.end(); ++i)
						count += (*i)->thread_info((*k)->comdat->retlines);

						(*k)->comdat->retval_ulong = count;
					}
				break;

				case SLEEP:
					globalstate = GLOBAL_PAUSED;
				break;

				case CLOSE:
					if(feedback)
					{
						wake_socketeer = false; //no necromancy

						//no talking to the dead...
						unwatch_all(inners, (*k)->swr);
						for(vector<channel*>::iterator sci = scripters.begin(); sci != scripters.end(); ++sci)
						(*sci)->script_nirvana((*k)->swr);

						//Kill, kill, kill!
						debug("[mixer] triggering reap of thread %i", (int)(*(*k)->comdat->thread));
						reap_socketeer((*k)->comdat->thread);
					}
				break;
				case ADD_IN:
					debug("[mixer] creating inchannel %zu", inners.size());
					errcode = add_input(inners, (*k)->strings[0], insem);
				break;
				case ADD_OUT:
					debug("[mixer] creating ouchannel %zu", outers.size());
					errcode = add_output(outers, (*k)->strings[0]);
				break;
				case REM_IN:
					debug("[mixer] removing inchannel %li", (*k)->ints[0]);
					errcode = remove_input(inners, outers, (*k)->ints[0]);
				break;
				case REM_OUT:
					debug("[mixer] removing outchannel %li", (*k)->ints[0]);
					errcode = remove_output(inners, outers, (*k)->ints[0]);
				break;

				case SAY:
					//suboptimal... should use list of pointers and let swr delete
					xdebug("[mixer] saying %s on socket_writer %p",(*k)->strings[0].c_str(), (void*) (*k)->swr);
					if((*k)->swr != NULL) (*k)->swr->write_copy("[saying] " + (*k)->strings[0]);
					else cerr << "[main] homeless message: " << (*k)->strings[0] << endl;
				break;
			}
			sxdebug("[mixer] end switch");
		}

		if( feedback && wake_socketeer )
		{
			(*k)->comdat->errcode = errcode;
			(*k)->comdat->retval = retval;
			// TODO: do this error text handling properly
			if(errtext != "") (*k)->comdat->errtext = errtext;
			sdebug("[mixer] notifying socketeer");
			sem_post((*k)->comdat->waitress);
		}
		// delete action if it was an archived script action
		if(aid != CLOSE && (*k)->comdat == NULL)
		{
			debug("deleting script action at %p of type %i", *k, (*k)->def->id); 
			delete *k;
		}
	}

	acts.clear();
}
