/*
	mixer_tools: tools for mixer
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"

#include "audio_buffer.hxx"
#include "audio_functions.hxx"

#include "input.hxx"
#include "output.hxx"

#include "netcomm.hxx"
#include "socket_writer_data.hxx"
#include "action.hxx"
#include "mixer_data.hxx"
#include "mixer_tools.hxx"

//#define DEBUG_TOOLS

string stringstatus(signed char status)
{
	string retro;
	switch(status)
	{
		case PLAYING:
			retro = "playing"; 
		break;
		case PAUSED:
			retro = "paused";
		break;
		case DEAD:
			retro = "dead";
		break;
		case IDLE:
			retro = "idle";
		break;
		case NODEV:
			retro = "nodev"; 
		break;
		case STOPPED:
			retro = "stopped";
		break;
		case SEEKING_PLAY:
			retro = "seek-play";
		break;
		case SEEKING_PAUSE:
			retro = "seek-pause";
		break;
		case LOADING:
			retro = "loading";
		break;
		default:
		{
			char sd[5]; sprintf(sd, "%d", status);
			retro = sd; 
		}
	}
	return retro;
}

string compatstatus(signed char status)
{
	switch(status)
	{
		case PAUSED:
		case SEEKING_PAUSE:
			return "stopped-pause";
		case PLAYING:
		case SEEKING_PLAY:
			return "playing";
		default:
			return "stopped";
	}
}

/*
	mp3
	raw:s16:2ch:44100Hz ... for signed short format with 2 channels and 44100Hz sampling freq.
	alias for 2ch is stereo, for 1ch mono
	raw_s16 is an alias for raw:s16:2ch:
*/
void audiodata::parse(const string &t)
{
	exists = true; /* also reliable? */
	string::size_type i = 0;
	string::size_type o = 0;
	/* parse tokens separated by : */
	while(i < t.length())
	{
		o = t.find(':', i);
		string::size_type l = (o == string::npos ? o : o-i);
		debug("found token %s", t.substr(i,l).c_str());
		if(i == 0) type = t.substr(i,l);
		else
		{
			string m = t.substr(i,o-i);
			if(m == "s16") sdebug("Got signed 16bit format... the only one I support atm.");
			else if(m == "stereo" || m == "2ch") format.channels = 2;
			else if(m == "mono" || m == "1ch") format.channels = 1;
			else if(m.length() > 2 && m.substr(m.length()-2) == "Hz") format.rate = strtoul(m.c_str(), NULL, 10);
			else
			{
				exists = false;
				problem = "parse error -- unknown token " + m;
				return;
			}
		}
		if(o == string::npos) break;
		i = o+1;
	}
}

/*
	Type
	Samples
	SampleRate
	Channels
	RVA
	Spezi 1 (MP3SamplesPerFrame)
*/

audiodata* get_audiodata(string sname, string type)
{
	char tline[20];
	audiodata* ad = new audiodata;
	if(sname.empty()) return ad;

	if(type != "auto")
	{
		ad->parse(type);
	}
	else
	{
	
	#ifdef DEBUG_TOOLS
	cerr << "get_audiodata: accessing file " << sname;
	#endif
	ifstream bfile(sname.c_str(), ios::in);
	if(bfile.is_open())
	{
		bfile.close();
		#ifdef DEBUG_TOOLS
		cerr << "successful\n";
		#endif
		ad->exists = true;
		
		string::size_type slashpos = sname.find_last_of("/",sname.length());
		if(slashpos == string::npos) slashpos = 0;
		sname = sname.substr(0, slashpos) + "/." + sname.substr(slashpos + 1) + ".audiodata"; 	
		
		#ifdef DEBUG_TOOLS
		cerr << "trying to open audiodata file " << sname << endl;
		#endif
		ifstream afile(sname.c_str(), ios::in);
		if(afile.is_open())
		{
			#ifdef DEBUG_TOOLS
			cerr << "opened " << sname << endl;
			#endif
			long int num = 0; //need to detect negative values as error!
			double dum = 0;
			int i = -1;
			while( afile.is_open() && (!afile.eof()) )
			{
				afile.getline(tline,19);
				//cerr << "read: " << tline << endl;
				
				if(afile.eof()) break;
				if(! ++i)
				{
					ad->type = tline;
				}	
				else
				{
					num = atoi(tline);
					dum = atof(tline);
				}
				//weak check for validity of data
				if((i == 1) && (num < 0)) break;
				//cerr << "got number " << num << endl; 
				switch(i)
				{
					case 0: break;
					case 1: ad->length = num; break;
					case 2: ad->format.rate = num; break;
					case 3: if(num < 10) ad->format.channels = num; break; //quite arbitary chosen... not nice
					case 4: ad->rva = dum; break; 
					default: ad->specials.push_back(num);
				}
			}
			if(i > 2) ad->reliable = true;
			else
			{
				//at least set length to sth no zero
				//the default for rate is prolly to arbitrary, too
				ad->length = ad->format.rate; //1 second
			}
			afile.close(); 
		}
		else
		{
			//try to guess at least type from file name
			//we have <unknown stuff of at least one character>.audiodata
			//so, rfind from position length() - 11
			
			//later the length should be estimated somehow (or read from decoder)
			ad->length = ad->format.rate; //1 second
			
			string::size_type i;
			//hack for mp3 radio stream
			//cerr <<  "search http:// in " << sname << endl;
			i = sname.find("http://",0);
			if(i == 0) ad->type = "mpeg";
			else
			{
				#ifdef DEBUG_TOOLS
				cerr << "get_audiodata searching in " << sname << endl;
				#endif
				string::size_type endingend = sname.length() - 11;
				i = sname.rfind(".", endingend);
				if(i != string::npos)
				{
					#ifdef DEBUG_TOOLS
					cerr << "ending in substr from " << i << "+1 to " << endingend << endl;
					#endif
					ad->type = sname.substr(i+1, endingend - i);
					#ifdef DEBUG_TOOLS
					cerr << "get_audiodata got type " << ad->type << endl;
					#endif
				}
			}
		}
	}
	else //could be url, assuming only mpeg streams - how to distinguish urls?
	{
		//in future one should read http playlists and parse to find real stream urls and format(s)
		string::size_type i = sname.find("http://",0);
		if(i == 0)
		{
			ad->exists = true;
			ad->length = ad->format.rate;
			ad->type = "mpeg";
		}
		else
		{
			i = sname.find("://",0);
			if((i != string::npos) && (i > 0))
			{
				ad->exists = true;
				ad->length = ad->format.rate;
				ad->type = sname.substr(0, i);
			}
			else
			{
				ad->problem = "invalid resource";
				serror("failed!");
			}
			sdebug("get_audiodata finished");
		}
	}
	}
	//MP3 and MP2 are both handles by the MPEG player
	if((strcasecmp(ad->type.c_str(), "mp3") == 0) || (strcasecmp(ad->type.c_str(), "mp2") == 0) || (strcasecmp(ad->type.c_str(), "mpg") == 0)) ad->type = "mpeg";
	/* hack for raw_s16 ... later be more flexible */
	else if(strcasecmp(ad->type.c_str(), "raw") == 0) ad->type = "raw_s16";

	sdebug("get_audiodata really finished");
	return ad;
}

// didn't know where else to put...
int sem_cancel(pthread_t thread, sem_t *sem)
{
	int r = pthread_cancel(thread);
	if(r == 0) sem_post(sem);
	return r;
}
