#ifndef DMD_H_NETCOMM
#define DMD_H_NETCOMM

/*
	netcomm: tcp port watching and basic socket communication
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
	derived from:
	
	netcomm.c for mixplayd
	
	provides functions for communiction on a tcp socket
	(c) Siegfried Wagner 01/06/27
	
	license: GNU General Public License (GPL)
	         For details see the file COPYING included with this archive
	last update: 01/08/28 by Siegfried Wagner
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

//beware of cancellation points... does capsulation of reading prevent this?

#define MAX_MSG_LENGTH 10000
#define MSG_BUF_SIZE 1024

class socket_buffer
{
	private:
	char buf[MSG_BUF_SIZE];
	int pos;
	int fill;
	
	public:
	int sock;
	string spyinprefix;
	string spyoutprefix;
	
	socket_buffer(int s);
	~socket_buffer();
	
	int read_line(string& msg); //doesn't include the "\n"!
	int write(string& msg); //doesn't automatically append "\n"
	int writeln(string* msg); //does automatically append "\r\n"
};

class socket_listener
{
protected:
	int sock;
public:
	bool online;
	socket_listener(void);
	virtual ~socket_listener();
	virtual int accept_connection(void);
};

class port_listener: public socket_listener
{
public:
	port_listener(porttype port_number, int local);
	~port_listener();
};

class unix_listener: public socket_listener
{
private:
	string sockname;
public:
	unix_listener(const string name);
	~unix_listener();
};

#endif
