/*
	out_oss: oss output device
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
*/

/*
	initially adapted from mixplayd 0.55... a long time ago
	
	Basically, I took the code and wrapped a c++ class around it.
	Of course it grew (and shrunk) to look quite different than before...
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "out_oss.hxx"
#include <sys/ioctl.h>
#include <fcntl.h>

#include <sys/soundcard.h>
#include "secsleep.hxx"

const string oss_output::t = "oss";

//for threaded variant
void* oss_writer(void* p)
{
	writerdata* data = (writerdata*) p;
	oss_writerextra* extra = (oss_writerextra*) data->extra;
	#ifdef DEBUG_OUTPUT
	cerr << "[oss_output] writer thread started" << endl; 
	#endif
	data->thread_workout(0);
	while(1)
	{
		safe_sem_wait(&data->insem);
		data->error = 0;
		#ifdef DEBUG_OUTPUT
		cerr << "[oss_output] writer thread writing" << endl; 
		#endif
		
		if(write(extra->fd,const_cast<audio_type*> (data->buffer.ptr),data->buffer.my_current_size()) == -1) data->error = 1;

		#ifdef DEBUG_OUTPUT
		cerr << "[oss_output] writer thread wrote" << endl; 
		#endif

		//sem_post(&data->outsem); //debugging!
		//secsleep((double) data->buffer.size/48000/10);
		sem_post(&data->outsem);
	}
	pthread_exit(NULL);
}

//for serial variant
void oss_output::special_write()
{
	wdat.error = 0;
	if(write(wextra.fd,const_cast<audio_type*> (wdat.buffer.ptr),wdat.buffer.my_current_size()) == -1) wdat.error = 1;
}

oss_output::oss_output(output_data* odi, unsigned int bufsize, const string type)
{
	construct("oss_output");
	current_type = type;
	if(type == "oss_serial") build(odi, bufsize, NULL);
	else if((type == "oss") || (type == "oss_threaded")) build(odi, bufsize, oss_writer);
	else
	{
		problem = "unknown variant " + type + " ... I only know oss, oss_threaded (aliased by oss) or oss_serial";
		prebuild(); //destroy would do havoc trying to destroy nonexistent semahores and thread otherwise
	}
}

oss_output::~oss_output()
{
	destbegin("oss_output");
	destroy();	
	destend("oss_output");
}

bool oss_output::special_init()
{
	wdat.extra = (void*) &wextra;
	return true;
}

void oss_output::special_exit()
{
	#ifdef DEBUG_MEM
	cerr << "[oss_output" << data->id << "] leaving with dsp fd " << wextra.fd << endl;
	#endif
}

bool oss_output::special_activate()
{
	#ifdef DEBUG_OUTPUT
	cerr << "[oss_output] activate with file " << data->filename << endl;
	#endif

	bool succ = false;

	oss_stereo = (data->format.channels == 2) ? 1 : 0;

	/* look for possible formats in sys/soundcard.h */
	oss_format = AFMT_S16_LE; //signed 16bit little endian
	
	if((wextra.fd=open(data->filename.c_str(),O_WRONLY,0)) != -1)
	{

		fcntl(wextra.fd, F_SETFD, FD_CLOEXEC); //close on exec (after forking) - sense???

		//I don't want duplex here... I want simple blocking output!
		//hm, how do I get it?
		//ioctl(wextra.fd, SNDCTL_DSP_SETDUPLEX, 0);
		//ioctl(wextra.fd, DSP_CAP_DUPLEX, 0);
		
		if(ioctl(wextra.fd,SNDCTL_DSP_SETFMT,&oss_format) != -1)
		{
			if(oss_format == AFMT_S16_LE) //OSS handbook indicates that ioctl may succeed but hardware still may not support the format
			{
				if(ioctl(wextra.fd,SNDCTL_DSP_STEREO,&oss_stereo) != -1)
				{
					if(ioctl(wextra.fd,SNDCTL_DSP_SPEED,&(data->format.rate)) != -1)
					{
						succ = true;
						#ifdef DEBUG_OUTPUT
						cerr << "[oss_output] having set sample rate to " << data->format.rate << endl; 
						#endif
					}
					else problem = "cannot set sample rate";
				}
				else problem = "cannot set mono/stereo mode";
			}
			else problem = "wrong format after ioctl";
		}
		else problem = "cannot set format";
	}
	else problem = "cannot open that device file";
	#ifdef DEBUG_OUTPUT
	cerr << "[oss_output] activate success: " << succ << " with fd " << wextra.fd << endl;
	#endif
	return succ;
}

void oss_output::special_deactivate()
{
	//OSS manual says that one doesn't have to (or should) do this explicitly
	//ioctl(fd, SNDCTL_DSP_RESET, 0);
	#ifdef DEBUG_MEM
	int cr = close(wextra.fd);
	cerr << "[oss_output" << data->id << "] close( " << wextra.fd << ") returned " << cr << endl;
	#else
	close(wextra.fd);
	#endif
}


//there are SNDCTL_DSP_SILENCE and SNDCTL_DSP_SKIP in OSS4...
//these could do what I want... when they wouldn't delete samples...

void oss_output::special_pause()
{
	#ifdef DEBUG_OUTPUT2
	cout << "output::pause" << endl;	
	#endif
	//dammed: this doesn't work ?
	//ioctl(fd, SNDCTL_DSP_POST, 0);
	//this?
	char b[4*576]; //4 bytes per sample
	for(int i = 0; i < 576; ++i)
	{
		b[i] = 0;
	}
	write(wextra.fd,b,576*4);
	//no...
}
