/*
	out_raw_s16: RAW stereo 16bit output device
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "out_raw_s16.hxx"

#include <fcntl.h>

const string raw_s16_output::t = "raw_s16";

void* raw_s16_writer(void* p)
{
	writerdata* data = (writerdata*) p;
	raw_s16_writerextra* extra = (raw_s16_writerextra*) data->extra;
	#ifdef DEBUG_OUTPUT
	cout << "[raw_s16_output] writer thread started" << endl; 
	#endif
	data->thread_workout(0);
	while(1)
	{
		safe_sem_wait(&data->insem);
		data->error = 0;
		
		if(write(extra->fd,const_cast<audio_type*> (data->buffer.ptr),data->buffer.my_current_size()) == -1) data->error = -1;
		
		sem_post(&data->outsem);
	}
	pthread_exit(NULL);
}

raw_s16_output::raw_s16_output(output_data* odi, unsigned int bufsize)
{
	construct("raw_s16_output");
	build(odi, bufsize, raw_s16_writer);
}

raw_s16_output::~raw_s16_output()
{
	destbegin("raw_s16_output");
	destroy();
	destend("raw_s16_output");
}

bool raw_s16_output::special_init()
{
	wdat.extra = (void*) &wextra;
	return true;
}

void raw_s16_output::special_exit()
{
	#ifdef DEBUG_MEM
	cout << "[raw_s16_output << " << data->id << "] leaving..." << endl;
	#endif
}

bool raw_s16_output::special_activate()
{
	#ifdef DEBUG_OUTPUT
	cout << "[raw_s16_output << " << data->id << "] activate" << endl; 
	#endif
	return((wextra.fd=open(data->filename.c_str(),O_WRONLY|O_CREAT|O_APPEND, 0666)) != -1);
}

void raw_s16_output::special_deactivate()
{
	#ifdef DEBUG_OUTPUT
	cout << "[raw_s16_output << " << data->id << "] deactivate" << endl; 
	#endif
	close(wextra.fd);
	//close device/file/connection
}

void raw_s16_output::special_pause()
{
	#ifdef DEBUG_OUTPUT
	cout << "[raw_s16_output << " << data->id << "] pause" << endl; 
	#endif
}
