#ifndef DMD_H_SPIES
#define DMD_H_SPIES
/*
	spies: list of communicators that are available for peer2peer communication.

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2

	Main feature is a read/write barrier to prevent bork up when a spy leaves the scene...
	Btw, there's no error checking here... also, because
	 a) STL vectors don't give errors, just exceptions (which I don't use in DerMixD, and of which I am afraid, esp. in a pthread).
	 b) What to do when safe_sem_wait fails???

	Note: This probably will soon see some refactoring with spies to have a common base (template) class.
*/
#include "common.hxx"
#include "action.hxx"
#include "socket_writer_data.hxx"

class one_peer
{
	public:
	socket_writer_data* data;
	string name;
	string description;
};

class peers
{
	private:
	pthread_mutex_t lock;
	sem_t readers;
	vector<one_peer> them;
	void writing();
	void done_writing();
	void reading(){ pthread_mutex_lock(&lock); sem_post(&readers); pthread_mutex_unlock(&lock); }
	void done_reading(){ safe_sem_wait(&readers); };

	public:
	peers();
	~peers();
	// Send a message... failing when there is no peer of that name.
	bool message(action &act);
	// List all peers, using the specified action for formatting.
	void list(action &act, socket_writer_data &wr); 
	// Add a new entry, error if name already taken.
	bool add(socket_writer_data* peerdata, const string &name, const string &descr);
	// Remove one entry...
	bool remove(const string &name);
	// Remove all peer entries for that socket
	void eliminate(socket_writer_data* data);
	void eliminate_all(); // ...
};

#endif
