#ifndef DMD_H_SOCKET_WRITER_DATA
#define DMD_H_SOCKET_WRITER_DATA

/*
	socket_writer_data: data structure and write methods acting as interface for socker_writer thread (that is defined elsewhere)
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
	
	depends on netcomm.h to provide actual socket reading/writing
*/

#include "netcomm.hxx"
#include "thread_container.hxx"

class spies;

class swm
{
	public:
	string *msg;
	bool spied;
	swm(string *ms, bool spy){ msg = ms; spied = spy; }
	~swm(){};
};

class socket_writer_data: public thread_data
{
	public:

	pthread_t thread;
	pthread_mutex_t lock;
	sem_t semo;

	bool is_end;

	vector<swm> messages;
	vector<swm> old_messages;

	socket_writer_data();
	~socket_writer_data();

	void write_copy(const string &msg);
	void write_copies(vector<string*> &msgs);
	void write(string *msg, bool spy=false);
	void writes(vector<string*> *msgs);
	// Add current messages to old_messages, scheduling for flushing.
	// This is threadsafe, using the lock.
	void prepare_flush();
	// Actually flush old_messages -- this is _not_ thread safe; use it in one thread only!
	void flush(socket_buffer* sb, spies *bnd = NULL);
	void end(); // Set the end marker, with the lock and semaphore to ensure the writer will be awake to get it.
};

// Implementation in spies.cxx .
class spies
{
	private:
	pthread_mutex_t lock;
	sem_t readers;
	vector<socket_writer_data*> them;
	void writing();
	void done_writing();

	public:
	volatile bool active;
	spies();
	~spies();
	// Call that before any read-only work (hand_over() is read-only).
	void reading(){ pthread_mutex_lock(&lock); sem_post(&readers); pthread_mutex_unlock(&lock); }
	// Give copies of msg to all spies, except the collaborateur himself.
	void hand_over(const string &msg, socket_writer_data *collaborateur);
	// Open up for changes of spies list again.
	void done_reading(){ safe_sem_wait(&readers); };
	// And these are the changes:
	void send(socket_writer_data* spy); // Add spy to spies list (if not already there).
	void eliminate(socket_writer_data* spy); // Remove the at most one occurence of that socket writer in spies list.
	void eliminate_all(); // ...
};

#endif
