#ifndef DMD_H_SPIES
#define DMD_H_SPIES
/*
	spies: list of communicators spying on all communication

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2

	Main feature is a read/write barrier to prevent bork up when a spy leaves the scene...
	Btw, there's no error checking here... also, because
	 a) STL vectors don't give errors, just exceptions (which I don't use in DerMixD, and of which I am afraid, esp. in a pthread).
	 b) What to do when safe_sem_wait fails???
*/
#include "common.hxx"
#include "socket_writer_data.hxx"

class spies
{
	private:
	pthread_mutex_t lock;
	sem_t readers;
	vector<socket_writer_data*> them;
	void writing();
	void done_writing();

	public:
	volatile bool active;
	spies();
	~spies();
	// Call that before any read-only work (hand_over() is read-only).
	void reading(){ pthread_mutex_lock(&lock); sem_post(&readers); pthread_mutex_unlock(&lock); }
	// Give copies of msg to all spies, except the collaborateur himself.
	void hand_over(const string &msg, socket_writer_data *collaborateur);
	// Open up for changes of spies list again.
	void done_reading(){ safe_sem_wait(&readers); };
	// And these are the changes:
	void send(socket_writer_data* spy); // Add spy to spies list (if not already there).
	void eliminate(socket_writer_data* spy); // Remove the at most one occurence of that socket writer in spies list.
	void eliminate_all(); // ...
};

#endif
