/*
	thread_container: something that contains threads

	part of DerMixD
	(c)2007-8 Thomas Orgis, licensed under GPLv2
*/
#include "thread_container.hxx"
#include <iostream>
#include <sstream>

void thread_data::thread_workout(int pri_offset)
{
#ifdef LINUX
	pid = syscall(__NR_gettid); // gettid();
	errno = 0;
	priority = getpriority(PRIO_PROCESS, pid);
	if(errno)
	{
		error("...getting priority: %s", strerror(errno));
		return;
	}
	fprintf(stderr, "Experiment with Linux/NTPL oddities: Shift my niceness by %i from %i (base=%i).\n",
	        pri_offset, priority, base_priority);
	if(setpriority(PRIO_PROCESS, 0, priority+pri_offset) < 0)
	error("Failed to set priority: %s", strerror(errno));
	else
	priority+=pri_offset;
#endif
}

threadspec::threadspec(){ construct("threadspec"); }

threadspec::threadspec(const pthread_t tid, const string ttext, thread_data *ttd)
{
	construct("threadspec");
	id = tid;
	text = ttext;
	td = ttd;
}

threadspec::~threadspec(){ destruct("threadspec"); }

void threadspec::add_info(vector<tstring*> &tl)
{
	ostringstream os; // Use the stream approach since id can be whatnot.
	os << id;
#ifdef LINUX
	if(td != NULL) os << "(" << td->pid << " @ " << td->priority << ")";
#endif
	os << ": " << text;
	tl.push_back(new tstring(os.str()));
}

thread_container::thread_container(){ construct("thread_container"); sub_thread_container = NULL; }
thread_container::~thread_container()
{
	destbegin("thread_container");
	for(vector<threadspec*>::iterator i = threads.begin(); i != threads.end(); ++i)
	delete (*i);

	destend("thread_container");
}

void thread_container::add_thread(const pthread_t id, const string text, thread_data *ttd)
{
	threads.push_back(new threadspec(id, text, ttd));
}

size_t thread_container::thread_info(vector<tstring*> &tl)
{
	size_t count = threads.size();
	for(vector<threadspec*>::iterator i = threads.begin(); i != threads.end(); ++i)
	(*i)->add_info(tl);

	if(sub_thread_container != NULL) count += sub_thread_container->thread_info(tl);
	return count;
}
