/*
	tstring: a C++ STL string that can do sprintf on itself

	part of DerMixD
	(c)2008 Thomas Orgis, licensed under GPLv2
*/
#include "tstring.hxx"
using namespace std;

// Chunks for increasing the sprintf buffer.
// On a sane libc, that is not even needed as snprintf will say how big the buffer needs to be.
#define CHUNK 50

tstring& tstring::sprintf(const char *fmt, ...)
{
	// Quick: Start out on the stack.
	char tsbuf_stack[50];
	char *tsbuf = tsbuf_stack;
	size_t tsbufs = 50;
	int fit = -1;
	do
	{
#ifdef DEBUG
		errno = 0;
#endif
		va_list ap;
		va_start(ap, fmt);
		fit = vsnprintf(tsbuf, tsbufs, fmt, ap);
		va_end(ap);
		if(fit >= (int)tsbufs || fit<0) // fit<0 is for old glibc and whoever.
		{
			tsbufs = fit > -1 ? fit+1 : tsbufs+CHUNK;
			debug("fit=%i, increasing size to %zu (error: %s)", fit, tsbufs, strerror(errno));
			// The full set of realloc paranoia.
			if(tsbuf == NULL || tsbuf == tsbuf_stack)
			{
				tsbuf = (char*)malloc(tsbufs);
				if(tsbuf == NULL) return *this;
			}
			else
			{
				char *tb = (char*)realloc(tsbuf, tsbufs);
				if(tb == NULL) // Should not happen...
				{
					free(tsbuf);
					tsbufs = 0;
					return *this;
				}
				tsbuf = tb;
			}
		}
		else break;
	} while(1);
	// Now append the sprintf'ed string.
	*this += tsbuf;
	if(tsbuf != tsbuf_stack) free(tsbuf);

	return *this;
}

void tstring_v_string(vector<tstring*> &tt, vector<string*> &st)
{
	for(vector<tstring*>::iterator s=tt.begin(); s != tt.end(); ++s)
	st.push_back(*s);
}
